from __future__ import annotations

import abc
from itertools import islice
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from roman_datamodels.datamodels import RampModel, RefpixRefModel

from dataclasses import dataclass
from enum import IntEnum

import numpy as np
from scipy import fft


class Const(IntEnum):
    """
    The necessary assumed values for the reference pixel correction
        - REF: the number of columns used by the detector reference pixels
        - PAD: the number of columns to add to introduce the appropriate delay
               between channel reads from the guide window scan and other processes
               this enables uniform time sampling of the data.
        - CHAN_WIDTH: the number of columns in a channel (detector amplifier)
        - N_CHANNELS: the number of channels for the detector
        - N_COLUMNS: the total number of columns in the detector
    """

    REF = 4
    PAD = 12
    CHAN_WIDTH = 128
    N_DETECT_CHAN = 32
    N_COLUMNS = CHAN_WIDTH * N_DETECT_CHAN


_offset = Optional[np.ndarray]


@dataclass
class BaseView(abc.ABC):
    """Base description for the data views"""

    data: np.ndarray
    offset: _offset = None

    @abc.abstractproperty
    def detector(self) -> np.ndarray:
        """View of the detector pixels."""
        ...

    @abc.abstractproperty
    def left(self) -> np.ndarray:
        """View of the left reference pixels."""
        ...

    @abc.abstractproperty
    def right(self) -> np.ndarray:
        """View of the right reference pixels."""
        ...

    @abc.abstractproperty
    def amp33(self) -> np.ndarray:
        """View of the amp33 reference pixels."""
        ...


@dataclass
class StandardView(BaseView):
    """
    The "standard" view for the computation, as in it is the one which most closely
    represents the data orientations found in reality. This view places the detector
    and amp33 pixels in the same array, with the detector pixel columns on the left
    and amp33 pixel columns on the right (as opposed to in a separate array).

    The single common array enables us to vectorize the computation and minimize
    data copies.

    The data dimensions are [frame, row, column], where the columns are given by
        - [detector, amp33]
        - left = detector[:Const.REF]
        - right = detector[-Const.REF:]
    """

    @classmethod
    def from_datamodel(cls, datamodel: RampModel) -> StandardView:
        """
        Read the datamodel into the standard view.
        """
        detector = datamodel.data

        # Extract amp33
        amp33 = datamodel.amp33
        # amp33 is normally a uint16, but this computation requires it to match
        # the data type of the detector pixels.
        amp33 = amp33.astype(detector.dtype)

        return cls(np.dstack([detector, amp33]))

    def update(self, datamodel: RampModel) -> RampModel:
        """
        Update the datamodel in place with the data from the standard view.
            - Returns the updated datamodel for a functional approach.
        """

        datamodel.data = self.detector
        # ABS to avoid casting negative numbers to uint16
        datamodel.amp33 = np.abs(self.amp33).astype(datamodel.amp33.dtype)
        datamodel.border_ref_pix_left = self.left
        datamodel.border_ref_pix_right = self.right

        return datamodel

    @property
    def detector(self) -> np.ndarray:
        return self.data[:, :, : Const.N_COLUMNS]

    @property
    def left(self) -> np.ndarray:
        return self.detector[:, :, : Const.REF]

    @property
    def right(self) -> np.ndarray:
        return self.detector[:, :, -Const.REF :]

    @property
    def amp33(self) -> np.ndarray:
        return self.data[:, :, -Const.CHAN_WIDTH :]

    @property
    def channels(self) -> ChannelView:
        """
        The data split by channel and padded for uniform time sampling.
            The resulting view's dimensions are [frame, row, channel, column], where the
            columns are padded by Const.PAD columns of zeros.

        Note this is NOT a view of the data, but a copy.
        """
        frames, rows, columns = self.data.shape
        channels = columns // Const.CHAN_WIDTH

        # First split data into channels
        #    [frame, row, channel, channel_column]
        # Second reorder so channels are first (better memory alignment)
        #    [channel, frame, row, channel_column]
        data = self.data.reshape((frames, rows, channels, Const.CHAN_WIDTH)).transpose(
            (2, 0, 1, 3)
        )

        # Reverse every other channel column so columns are in reading order this is a
        # view of the data, not a copy at this point so this will back-propagate to the
        # data stored in the class.
        # This will need to be undone before returning the data because otherwise
        # the next call to this method will be different (columns will be erroneously
        # reversed a second time).
        # The next step of this method will pad the data, breaking the view, so we
        # can safely undo the changes afterwords.
        # We do not make an original copy of from the data so that we avoid having to
        # copy the data twice.
        # Maybe this should be done with a context manager?
        data[1::2, :, :, :] = data[1::2, :, :, ::-1]

        # pad channels with zeros to account for the pause at the end of each read
        # delay return until undoing the column reversal
        output = np.pad(
            data, ((0, 0), (0, 0), (0, 0), (0, Const.PAD)), constant_values=0
        )

        # Undo the column reversal
        data[1::2, :, :, :] = data[1::2, :, :, ::-1]

        return ChannelView(output, offset=self.offset)

    def remove_offset(self) -> StandardView:
        """
        Use linear least squares regression to compute the best fit intercept
        (offset) of all the data and then subtract that from the data.
            - This records the offset in the class so that it can be added back
              to the data later.
            - Returns the class back to the user even though it will be modified
              in place by the views. This is so that it can be treated functionally.
        """

        frames, rows, columns = self.data.shape

        # Reshape data so that it is a view of the data of shape:
        #    [frame, frame_data]
        # where frame_data is the data for a single frame
        data = self.data.reshape((frames, rows * columns))

        # Craate an independent variable indexed by frame and centered at zero
        indep = np.arange(frames, dtype=data.dtype)
        indep = indep - np.mean(indep)

        # Compute sums needed for linear least squares
        sx = np.sum(indep)
        sxx = np.sum(indep**2)
        sy = np.sum(data, axis=0)
        sxy = np.matmul(indep, data, dtype=data.dtype)

        # Compute the offset (y-intercept) for the fit
        offset = (sy * sxx - sx * sxy) / (frames * sxx - sx**2)

        # Apply the offset to the data (in place)
        data -= offset

        self.offset = offset.reshape(rows, columns)

        return self

    def apply_offset(self) -> StandardView:
        """
        Apply the recorded offset in-place and return the class to the user.
            - If the offset has not been recorded, this will do nothing and return the
              object.
            - Returns the class back to the user even though it will be modified
              in place by the views. This is so that it can be treated functionally.
        """

        if self.offset is not None:
            self.data += self.offset
            self.offset = None

        return self


class ChannelView(BaseView):
    """
    This is the "channel" view of the data, which happens to be the most convenient
    view for performing the correction's computations.

    The actual data storage dimensions are
        [channel, frame, row, column],
    where the columns are the in-channel columns of the data.
        - The odd index channels column order is reversed relative to the appearance
          in the StandardView.
        - The padding columns are added to the end of each set of channel columns
          AFTER the reversal of the odd index channels.

    The odd index channels are reversed so that the columns appear aligned in the
    read order of the detector, so that when we compute transforms on the flattened
    data of each channel for each frame, the data is ordered in the order it was
    read in time.

    The padding columns are added to account for the exact delay between column
    reads (it would have read that many pixes in the time it takes to move to the
    next frame read).
    """

    @property
    def detector(self) -> np.ndarray:
        return self.data[: Const.N_DETECT_CHAN, :, :, :]

    @property
    def left(self) -> np.ndarray:
        """
        Left reference pixels, zero padded so that it has 140 columns to match the
        amp33 pixels.

        This is not a view, but a copy so that the data is not modified by the
        operation.
        """
        left = np.zeros(self.detector.shape[1:], dtype=self.detector.dtype)
        left[:, :, :4] = self.detector[0, :, :, :4]

        return left

    @property
    def right(self) -> np.ndarray:
        """
        Right reference pixels, zero padded so that it has 140 columns to match the
        amp33 pixels.

        This is not a view, but a copy so that the data is not modified by the
        operation.

        Note that due to channel flipping, the right reference pixels are the first
        four columns of the last detector channel. (backwards to initial guess)
        """
        right = np.zeros(self.detector.shape[1:], dtype=self.detector.dtype)
        right[:, :, :4] = self.detector[-1, :, :, :4]

        return right

    @property
    def amp33(self) -> np.ndarray:
        """
        amp33 reference pixels

        Note that this returns a view of the data, not a copy. This is to make
        some of the smoothing and interpolations occur in place.
        """

        return self.data[-1, :, :, :]

    @property
    def standard(self) -> StandardView:
        """
        Transform the data back into the standard view.
            - Removes the padding columns
            - Reverses the flipped odd index channels.
        """

        # Remove the padding from the data, copy is to break the view back to the
        # original data, this will be the only copy of the data needed.
        data = self.data[:, :, :, : -Const.PAD].copy()
        channels, frames, rows, columns = data.shape

        # Reverse every other channel column so columns are in standard order
        data[1::2, :, :, :] = data[1::2, :, :, ::-1]

        # Undo channel reordering to:
        #    [frame, row, channel, channel_column]
        # Second combine the channels into columns:
        #    [frame, row, column]
        data = data.transpose((1, 2, 0, 3)).reshape((frames, rows, columns * channels))

        return StandardView(data, offset=self.offset)

    def remove_trends(self) -> ChannelView:
        """
        Remove the linear in time (by frame) trends from the boundary data.
            This is different than the offset correction, which is computed
            simultaneously across all frames in order to remove any general bias.

        Note:
            - Change will be inplace of this object.
            - It will also return this object so it can be treated functionally.
        """
        _, _, rows, columns = self.data.shape

        # Create an independent variable indexed by frame and centered at zero
        t = np.arange(columns * rows, dtype=self.data.dtype).reshape((rows, columns))
        t = t - np.mean(t)

        # Locate the top and bottom reference pixel rows
        REF_ROWS = [*np.arange(Const.REF), *(rows - np.arange(Const.REF) - 1)[::-1]]

        # Restrict to the reference pixels and non-zero values
        t_ref = t[REF_ROWS, :]
        not_zero = self.data != 0

        # fit needs to be done for each channel and frame separately
        for chan_data, chan_not_zero in zip(self.data, not_zero):
            for frame_data, frame_not_zero in zip(chan_data, chan_not_zero):
                # Find the non-zero reference pixels for this channel and frame
                mask = frame_not_zero[REF_ROWS, :]

                # Compute the independent and dependent variables for the fit
                x_vals = t_ref[mask]
                y_vals = frame_data[REF_ROWS, :][mask]

                # Skip if there is no data
                if x_vals.size < 1 or y_vals.size < 1:
                    continue

                # Perform the fit using a 1st order polynomial, (i.e. linear fit)
                m, b = np.polyfit(x_vals, y_vals, 1)

                # Remove the fit from the data
                frame_data -= (t * m + b) * frame_not_zero

        return self

    def cosine_interpolate(self) -> ChannelView:
        """
        Perform Cosine weighted interpolation on the zero values of the amp33
        channel.
            - This is done in place.
            - This returns the object so it can be used functionally.

        I have purposely let this computation be verbose because we may have to
        apply it to both the left and right reference pixel channels.
        """
        data = self.amp33
        channels, _, rows, columns = self.detector.shape

        # Create the "cosine" interpolation
        interp = np.sin(np.arange(1, channels + 1, dtype=data.dtype) * np.pi / channels)

        for frame in data:
            # flatten the frame into a 1D array
            kern = frame.reshape(rows * columns)

            # Find the non-zero values (so we can normalize by their values later)
            #    Note: I do not thing the dtype cast is necessary here, but it was
            #          specifically cast this way in the original code. I suspect
            #          we can remove the dtype cast here to avoid a data copy. For
            #          little loss of numerical precision and a modest speedup.
            mask = (kern != 0).astype(np.int64)

            # Convolve the interpolation with the data
            cov = np.convolve(kern, interp, mode="same")
            mask_conv = np.convolve(mask, interp, mode="same")

            # Normalize by the non-zero values
            kern = (cov / mask_conv).reshape(rows, columns)

            # Apply the interpolation to the data
            frame += kern * (1 - mask.reshape(rows, columns))

        # Fix NaN values because the above computations can interpolate to NaN
        #     values, these will be set to zero.
        np.nan_to_num(data, copy=False)

        return self

    @staticmethod
    def fft_interp_generator(
        frame: np.ndarray, pad: np.ndarray, apodize: np.ndarray
    ) -> np.ndarray:
        """
        Provide an infinite generator of FFT interpolation iterations using the
        apodizing function for the padded columns of a given frame.

            Since this is an iterative method, and determining the number of iterations
            is arbitrary, a generator is provided so tha it is simple to choose method
            of stopping the iteration (fixed, relative change, absolute change, etc.)
        """
        while True:
            # Perform the FFT interpolation using the apodizing function
            result = apodize * fft.rfft(frame, workers=1) / frame.size
            result = fft.irfft(result * frame.size, workers=1).astype(frame.dtype)

            # Only update the padded columns
            frame[pad] = result[pad]

            yield frame

    def fft_interpolate(self, num: int = 3) -> ChannelView:
        """
        FFT interpolate the amp33 reference channel's padded columns.
            - This is done in place.
            - The object is returned so it can be used functionally.

        Again, I have purposely let this computation be verbose because we may
        have to apply it to both the left and right reference pixel channels.

        Parameters:
        ----------
            num: The number of iterations to perform. (default: 3)
        """
        frames, rows, columns = self.amp33.shape
        length = rows * columns

        # Mask all the data columns
        mask = np.ones((rows, columns), dtype=bool)
        mask[:, : -Const.PAD] = (
            False  # this maybe different for the left/right channels
        )
        mask = mask.flatten()

        # Find the indices of the padded columns
        pad = np.where(mask)[0]

        data = self.amp33.reshape(frames, length)
        apodize = (
            1 + np.cos(2 * np.pi * np.abs(np.fft.rfftfreq(length, 1 / length)) / length)
        ) / 2

        for frame in data:
            # Note `islice(iter, start, stop)` is tool which takes in an iterator
            # and returns selected elements from it.
            #   - `start` is the index of the first element to return from the iterator
            #   - `stop` is the index of the last element to return, if it is None
            #      then an iterator starting at `start` is returned.
            #
            # In this case we are using it to advance (skip) the generator (returns an
            # iterator) to the  `num - 1` index term (i.e. run the generator `num`
            # times). Since`stop = None`, it will return an iterator whose first
            # output will be the `num`th iteration of the algorithm (the one we want).
            #
            # `next(iter)` will return the next value from the iterator, then advance
            #              the iterator.
            #    - So running it once on an iterator will return the first value
            #      of the iterator. Running again on the iterator will return the next
            #      value, and so on.
            #
            # In this case, we have the infinite iterable, whose first value is
            # the iteration of the algorithm we want from `islice`, so calling
            # `next()` on it will simply return the iteration we desire.
            #
            # Other itertools can be use to recreate any iteration termination behavior
            # desired. In this case, we put them together to iterate a fixed number of
            # times.

            next(
                islice(
                    self.fft_interp_generator(frame, pad, apodize),
                    num - 1,
                    None,
                )
            )

        return self

    @staticmethod
    def forward_fft(channel: np.ndarray, normalize: bool) -> np.ndarray:
        """
        Compute the forward FFT of the channel, and normalize if requested.
            - Normalization is required when the data is zero-padded. In
              the left/right channels, the only Const.REF columns are
              non-zero. So we need to normalize by the number of
              columns / Const.REF.
        """
        frames, rows, columns = channel.shape
        channel = channel.reshape(frames, rows * columns)

        channel = fft.rfft(channel / channel.shape[1])
        if normalize:
            channel *= columns / Const.REF

        return channel

    @property
    def reference_fft(self) -> ReferenceFFT:
        """
        Compute the FFTs of all the reference pixels.
        """
        return ReferenceFFT(
            self.forward_fft(self.left, True),
            self.forward_fft(self.right, True),
            self.forward_fft(self.amp33, False),
        )

    def correction(self, coeffs: Coefficients) -> np.ndarray:
        """
        Compute the the correction array for the the detector.
        """
        correction = self.reference_fft.correction(coeffs)

        return correction.reshape(self.data.shape).astype(self.data.dtype)

    def apply_correction(self, coeffs: Coefficients) -> StandardView:
        """
        Apply the correction and return the standard view of the data
            - This is done in place.
            - The standard view is returned because that is what we ultimately
              want.
        """
        self.data -= self.correction(coeffs)

        return self.standard


@dataclass
class ReferenceFFT:
    """
    FFTs of the Reference Pixels.

    Note that the FFTs have all been reshaped so that the rows and columns are
    combined into a single array. This means the FFTs are of shape
        - (frames, (rows * columns) // 2 + 1).

    where the // 2 + 1 is the length of the FFT for real data of even length
    """

    left: np.ndarray
    right: np.ndarray
    amp33: np.ndarray

    def channel_correction(self, coeffs: Coefficients) -> np.ndarray:
        """
        Generator which yields the correction for each channel.
            This can then be stacked into a single array to form the correction.
        """
        # We need to multiply by 2 because we are only using half of the FFT because
        # the data is real
        normalization = coeffs.gamma.shape[1] * 2

        for gamma, zeta, alpha in coeffs:
            correction = (
                np.multiply(self.left, gamma)
                + np.multiply(self.right, zeta)
                + np.multiply(self.amp33, alpha)
            ) * normalization

            # hold onto the previous correction so that shape is maintained
            # for the blank correction for the amp33 channel
            correction = fft.irfft(correction)
            yield correction

        # Add zeros in for the amp33 channel as it does not get changed
        yield np.zeros(correction.shape)

    def correction(self, coeffs: Coefficients) -> np.ndarray:
        """
        Get the correction array for all of the data.
            - Stacks all channels into a single array

        Returns the correction array with the rows and columns in a single combined
        dimension, and the dtypes not adjusted to the original dtypes.
            - This will be handled by the ChannelView class
        """

        return np.array(list(self.channel_correction(coeffs)))


@dataclass
class Coefficients:
    """
    Convenience class for holding the coefficients from the reference file.

    It holds references to them, so it is not a copy and provides a simple
    interface for iterating over them.
    """

    gamma: np.ndarray
    zeta: np.ndarray
    alpha: np.ndarray

    @classmethod
    def from_ref(cls, ref: RefpixRefModel) -> Coefficients:
        return cls(ref.gamma, ref.zeta, ref.alpha)

    def __iter__(self):
        return zip(self.gamma, self.zeta, self.alpha)



from astropy.table import Table


def get_photometry_columns(table):
    """
    Get the photometry-related column names in an astropy table.

    Parameters
    ----------
    table : astropy.table.Table
        The table to search for photometry-related columns.

    Returns
    -------
    result: list
        The list of photometry-related column names.
    """
    if not isinstance(table, Table):
        raise ValueError("table must be an astropy.table.Table object")

    phot_cols = []
    for col in table.colnames:
        if (
            col.startswith("aper")
            or col.startswith("CI_")
            or col.startswith("isophotal_flux")
            or col.startswith("kron_flux")
            or "_psf" in col
        ):
            phot_cols.append(col)

    return phot_cols


def prefix_colnames(table, prefix):
    """
    Update column names in an astropy table in-place by adding a prefix
    to photometry-related columns.

    Parameters
    ----------
    table : `~astropy.table.Table`
        The table to update.

    prefix : str
        The prefix to add to the photometry-related column names. If
        "det_", the columns will be removed.

    Returns
    -------
    result : `~astropy.table.Table`
        The updated table.
    """
    if not isinstance(prefix, str):
        raise ValueError("prefix must be a string")

    phot_cols = get_photometry_columns(table)
    for col in table.colnames:
        if col in phot_cols:
            table.rename_column(col, prefix + col)

    return table


def remove_columns(table):
    """
    Remove photometry-related columns from an astropy table in-place.

    Parameters
    ----------
    table : `~astropy.table.Table`
        The table to update.

    Returns
    -------
    result : `~astropy.table.Table`
        The updated table.
    """
    phot_cols = get_photometry_columns(table)
    for col in table.colnames:
        if col in phot_cols:
            table.remove_column(col)

    return table



"""Update path of members in an association"""

from os.path import basename, join


def update_path(asn, file_path, target="expname"):
    """Update path of members in an association

    Parameters
    ----------
    asn : Association
        An association. The association is modified in-place.

    file_path : str
        New path to prepend to each member

    target : str
        Key to replace
    """
    update_key_value(asn, target, (file_path,), mod_func=_replace_path)


def update_key_value(obj, target, func_args, mod_func=None):
    """Update all instances of a key using a modifier

    Parameters
    ----------
    obj : object
        Hashable object to modify

    target : str
        The target key to modify

    func_args : (arg(, ...))
        Arguments to pass to the modification function

    mod_func : function
        Function to modify the target key with. If `None`,
        the key will be replaced by the arg list.

    Notes
    -----
    The first argument to `mode_func` will always be the value of the
    target key. Any other arguments given will then be passed to the
    the function.
    """
    if mod_func is None:

        def mod_func(value, args):
            return args

    if hasattr(obj, "items"):
        for key, value in obj.items():
            if key == target:
                obj[key] = mod_func(value, *func_args)
            if isinstance(value, dict):
                update_key_value(value, target, func_args, mod_func=mod_func)
            elif isinstance(value, list):
                for item in value:
                    update_key_value(item, target, func_args, mod_func=mod_func)


def _replace_path(old_path, new_path):
    """Replace the path prefix of the basename

    Parameters
    ----------
    old_path : str
        A string with a file path.

    new_path : str
        The new path to prepend to the basename.

    Returns
    -------
    new_full_path : str
        The basename of the path with the
    """
    file_name = basename(old_path)
    new_full_path = join(new_path, file_name)
    return new_full_path


# #####
# Tests
# #####
_test_obj = {
    "a": "change",
    "b": "nochange",
    "c": {"a": "change", "b": "nochange"},
    "d": [
        {"a": "change", "b": "nochange"},
        {"a": "change", "b": "nochange"},
    ],
}


def test_update_key_value_default():
    from copy import deepcopy

    obj = deepcopy(_test_obj)
    target = "a"
    new_value = "changed"
    update_key_value(obj, target, (new_value,))
    for value in _gen_dict_extract(target, obj):
        assert value == new_value


def test_update_key_value_mod_func():
    from copy import deepcopy

    obj = deepcopy(_test_obj)
    target = "a"
    new_value = "changed"

    def mod_func(v, suffix):
        return v + suffix

    update_key_value(obj, target, (new_value,), mod_func=mod_func)
    assert_values = [
        mod_func(value, new_value) for value in _gen_dict_extract(target, _test_obj)
    ]
    for value in _gen_dict_extract(target, obj):
        assert assert_values.pop(0) == value


def test_replace_path():
    new_path = join("hello", "there")
    base = "base.me"
    assert_value = join(new_path, base)
    assert assert_value == _replace_path(base, new_path)
    assert assert_value == _replace_path(join("somepath", base), new_path)


def _gen_dict_extract(key, var):
    if hasattr(var, "items"):
        for k, v in var.items():
            if k == key:
                yield v
            if isinstance(v, dict):
                yield from _gen_dict_extract(key, v)
            elif isinstance(v, list):
                for d in v:
                    yield from _gen_dict_extract(key, d)



#! /usr/bin/env python
#
import logging

import numpy as np
from astropy import units as u
from roman_datamodels import datamodels as rdd
from roman_datamodels import maker_utils
from roman_datamodels import stnode as rds
from roman_datamodels.dqflags import group, pixel
from stcal.ramp_fitting import ols_cas22_fit
from stcal.ramp_fitting.ols_cas22 import Parameter, Variance

from romancal.stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["RampFitStep"]


class RampFitStep(RomanStep):
    """
    This step fits a straight line to the value of counts vs. time to
    determine the mean count rate for each pixel.
    """

    class_alias = "ramp_fit"

    spec = """
        algorithm = option('ols_cas22', default='ols_cas22')  # Algorithm to use to fit.
        save_opt = boolean(default=False) # Save optional output
        suffix = string(default='rampfit')  # Default suffix of results
        use_ramp_jump_detection = boolean(default=True) # Use jump detection during ramp fitting
        threshold_intercept = float(default=None) # Override the intercept parameter for the threshold function in the jump detection algorithm.
        threshold_constant = float(default=None) # Override the constant parameter for the threshold function in the jump detection algorithm.
    """  # noqa: E501

    weighting = "optimal"  # Only weighting allowed for OLS

    reference_file_types = ["readnoise", "gain", "dark"]

    def process(self, input):
        with rdd.open(input, mode="rw") as input_model:
            # Retrieve reference info
            readnoise_filename = self.get_reference_file(input_model, "readnoise")
            gain_filename = self.get_reference_file(input_model, "gain")
            log.info("Using READNOISE reference file: %s", readnoise_filename)
            readnoise_model = rdd.open(readnoise_filename, mode="r")
            log.info("Using GAIN reference file: %s", gain_filename)
            gain_model = rdd.open(gain_filename, mode="r")

            # Do the fitting.
            algorithm = self.algorithm.lower()
            if algorithm == "ols_cas22":
                dark_filename = self.get_reference_file(input_model, "dark")
                dark_model = rdd.open(dark_filename, mode="r")
                out_model = self.ols_cas22(
                    input_model, readnoise_model, gain_model, dark_model
                )
                out_model.meta.cal_step.ramp_fit = "COMPLETE"
            else:
                log.error("Algorithm %s is not supported. Skipping step.")
                out_model = input
                out_model.meta.cal_step.ramp_fit = "SKIPPED"

        return out_model

    def ols_cas22(self, input_model, readnoise_model, gain_model, dark_model):
        """Peform Optimal Linear Fitting on arbitrarily space resulants

        Parameters
        ----------
        input_model : RampModel
            Model containing ramps.

        readnoise_model : ReadnoiseRefModel
            Model with the read noise reference information.

        gain_model : GainRefModel
            Model with the gain reference information.

        dark_model : DarkRefModel
            Model with the dark reference information

        Returns
        -------
        out_model : ImageModel
            Model containing a count-rate image.
        """
        use_jump = self.use_ramp_jump_detection

        if use_jump:
            log.info("Jump detection as part of ramp fitting is enabled.")
        else:
            log.info("Jump detection as part of ramp fitting is disabled.")

        kwargs = {}
        if self.threshold_intercept is not None:
            kwargs["threshold_intercept"] = self.threshold_intercept
        if self.threshold_constant is not None:
            kwargs["threshold_constant"] = self.threshold_constant

        resultants = input_model.data
        dq = input_model.groupdq
        read_noise = readnoise_model.data.value
        gain = gain_model.data.value
        read_time = input_model.meta.exposure.frame_time

        # Force read pattern to be pure lists not LNodes
        read_pattern = [list(reads) for reads in input_model.meta.exposure.read_pattern]
        if len(read_pattern) != resultants.shape[0]:
            raise RuntimeError("mismatch between resultants shape and " "read_pattern.")

        # add dark current back into resultants so that Poisson noise is
        # properly accounted for
        tbar = np.array([np.mean(reads) * read_time for reads in read_pattern])
        resultants += (
            dark_model.dark_slope.to(u.DN / u.s).value[None, ...] * tbar[:, None, None]
        )

        # account for the gain
        resultants *= gain
        read_noise *= gain

        # Fit the ramps
        output = ols_cas22_fit.fit_ramps_casertano(
            resultants,
            dq,
            read_noise,
            read_time,
            read_pattern=read_pattern,
            use_jump=use_jump,
            **kwargs,
        )

        # Break out the information and fix units
        slopes = output.parameters[..., Parameter.slope]
        var_rnoise = output.variances[..., Variance.read_var]
        var_poisson = output.variances[..., Variance.poisson_var]
        err = np.sqrt(var_poisson + var_rnoise)
        dq = output.dq.astype(np.uint32)

        # remove dark current contribution to slopes
        slopes -= dark_model.dark_slope.to(u.DN / u.s).value * gain

        # Propagate DQ flags forward.
        ramp_dq = get_pixeldq_flags(dq, input_model.pixeldq, slopes, err, gain)

        # Create the image model
        image_info = (slopes, ramp_dq, var_poisson, var_rnoise, err)
        image_model = create_image_model(input_model, image_info)

        # Rescale by the gain back to DN/s
        image_model.data /= gain[4:-4, 4:-4]
        image_model.err /= gain[4:-4, 4:-4]
        image_model.var_poisson /= gain[4:-4, 4:-4] ** 2
        image_model.var_rnoise /= gain[4:-4, 4:-4] ** 2

        # That's all folks
        return image_model


# #########
# Utilities
# #########
def create_image_model(input_model, image_info):
    """
    Creates an ImageModel from the computed arrays from ramp_fit.

    Parameters
    ----------
    input_model : `~roman_datamodels.datamodels.RampModel`
        Input ``RampModel`` for which the output ``ImageModel`` is created.

    image_info : tuple
        The ramp fitting arrays needed for the ``ImageModel``.

    Returns
    -------
    out_model : `~roman_datamodels.datamodels.ImageModel`
        The output ``ImageModel`` to be returned from the ramp fit step.
    """
    data, dq, var_poisson, var_rnoise, err = image_info

    if dq is None:
        dq = np.zeros(data.shape, dtype="u4")

    # Create output datamodel
    # ... and add all keys from input
    meta = dict(wcs=None)  # default empty WCS
    meta.update(input_model.meta)
    meta["cal_step"]["ramp_fit"] = "INCOMPLETE"
    meta["cal_logs"] = maker_utils.mk_cal_logs()
    meta["photometry"] = maker_utils.mk_photometry()
    inst = {
        "meta": meta,
        "data": data,
        "dq": dq,
        "var_poisson": var_poisson,
        "var_rnoise": var_rnoise,
        "err": err,
        "amp33": input_model.amp33.copy(),
        "border_ref_pix_left": input_model.border_ref_pix_left.copy(),
        "border_ref_pix_right": input_model.border_ref_pix_right.copy(),
        "border_ref_pix_top": input_model.border_ref_pix_top.copy(),
        "border_ref_pix_bottom": input_model.border_ref_pix_bottom.copy(),
        "dq_border_ref_pix_left": input_model.dq_border_ref_pix_left.copy(),
        "dq_border_ref_pix_right": input_model.dq_border_ref_pix_right.copy(),
        "dq_border_ref_pix_top": input_model.dq_border_ref_pix_top.copy(),
        "dq_border_ref_pix_bottom": input_model.dq_border_ref_pix_bottom.copy(),
    }
    out_node = rds.WfiImage(inst)
    im = rdd.ImageModel(out_node)

    # trim off border reference pixels from science data, dq, err
    # and var_poisson/var_rnoise
    im.data = im.data[4:-4, 4:-4]
    im.dq = im.dq[4:-4, 4:-4]
    im.err = im.err[4:-4, 4:-4]
    im.var_poisson = im.var_poisson[4:-4, 4:-4]
    im.var_rnoise = im.var_rnoise[4:-4, 4:-4]

    return im


def get_pixeldq_flags(groupdq, pixeldq, slopes, err, gain):
    """
    Construct pixeldq for ramp fit output from input dqs and ramp slopes.

    The algorithm is:
    - pass forward existing pixeldq flags
    - if we flagged a jump, flag the pixel as containing a jump
    - if everything is saturated, flag the pixel as saturated
    - if everything is saturated or do not use, flag the pixel as do not use
    - add NO_GAIN_VALUE if gain is not finite or less than zero

    Parameters
    ----------
    groupdq : np.ndarray
        dq flags for each resultant
    pixeldq : np.ndarray
        dq flags for each pixel
    slopes : np.ndarray
        derived slopes for each pixel
    err : np.ndarray
        derived total uncertainty for each pixel
    gain : np.ndarray
        gains for each pixel

    Returns
    -------
    pixeldq : np.ndarray
        Updated pixeldq array combining information from input dq and slopes.
    """
    outpixeldq = pixeldq.copy()
    # jump flagging
    m = np.any(groupdq & group.JUMP_DET, axis=0)
    outpixeldq |= (m * pixel.JUMP_DET).astype(np.uint32)
    # all saturated flagging
    m = np.all(groupdq & group.SATURATED, axis=0)
    outpixeldq |= (m * pixel.SATURATED).astype(np.uint32)
    # all either saturated or do not use or NaN slope flagging
    satordnu = group.SATURATED | group.DO_NOT_USE
    m = np.all(groupdq & satordnu, axis=0)
    m |= ~np.isfinite(slopes) | (err <= 0)
    outpixeldq |= (m * pixel.DO_NOT_USE).astype(np.uint32)
    m = (gain < 0) | ~np.isfinite(gain)
    outpixeldq |= (m * pixel.NO_GAIN_VALUE).astype(np.uint32)

    return outpixeldq



""" Roman tests for source catalog creation """

import asdf
import pytest

from romancal.stpipe import RomanStep

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(
    scope="module",
    params=[
        "r0099101001001001001_r274dp63x31y81_prompt_F158_i2d.asdf",
        "r0099101001001001001_F158_visit_i2d.asdf",
        "r0000101001001001001_0001_wfi01_cal.asdf",
    ],
    ids=["L3", "L2", "L3skycell"],
)
def run_source_catalog(rtdata_module, request):
    rtdata = rtdata_module

    inputfn = request.param

    outputfn = inputfn.rsplit("_", 1)[0] + "_cat.asdf"
    rtdata.output = outputfn

    rtdata.get_data(f"WFI/image/{inputfn}")
    rtdata.input = inputfn
    rtdata.get_truth(f"truth/WFI/image/{outputfn}")

    args = [
        "romancal.step.SourceCatalogStep",
        rtdata.input,
    ]
    RomanStep.from_cmdline(args)
    return rtdata_module


@pytest.fixture(scope="module")
def catalog(run_source_catalog):
    with asdf.open(run_source_catalog.output) as af:
        yield af["roman"]["source_catalog"]


@pytest.fixture(scope="module")
def fields(catalog):
    return catalog.dtype.names


@pytest.mark.parametrize(
    "field",
    (
        "ra_centroid",  # DMS374 positions on ICRF
        "dec_centroid",  # DMS374 positions on ICRF
        "aper30_flux",  # DMS399 aperture fluxes
        "aper50_flux",  # DMS399 aperture fluxes
        "aper70_flux",  # DMS399 aperture fluxes
        "aper_total_flux",  # DMS375 fluxes
        "is_extended",  # DMS376 type of source
        "aper_total_flux_err",  # DMS386 flux uncertainties
        "flags",  # DMS387 dq_flags
    ),
)
def test_has_field(fields, field):
    assert field in fields



"""Association Rules: Basic
"""

from romancal.associations import Association
from romancal.associations.lib.constraint import ConstraintTrue
from romancal.associations.registry import RegistryMarker


@RegistryMarker.rule
class Rule_1(Association):
    """Basic rule"""

    def __init__(self, version_id=None):
        self.constraints = ConstraintTrue()
        super().__init__(version_id=version_id)
        self.data["members"] = []

    def _add(self, item):
        self.data["members"].append(item)

    def is_item_member(self, item):
        """Check if item is already a member of this association

        Parameters
        ----------
        item: dict
            The item to add.

        Returns
        -------
        is_item_member: bool
            True if item is a member.
        """
        return item in self.data["members"]



import logging
from timeit import default_timer as timer

from ..lib.progress import Bar
from .association import make_timestamp
from .lib.process_list import (
    ListCategory,
    ProcessList,
    ProcessQueueSorted,
    workover_filter,
)
from .pool import PoolRow

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = ["generate"]


def generate(pool, rules, version_id=None, finalize=True):
    """Generate associations in the pool according to the rules.

    Parameters
    ----------
    pool : AssociationPool
        The pool to generate from.

    rules : AssociationRegistry
        The association rule set.

    version_id : None, True, or str
        The string to use to tag associations and products.
        If None, no tagging occurs.
        If True, use a timestamp
        If a string, the string.

    Returns
    -------
    associations : [Association[,...]]
        List of associations

    Notes
    -----
    Refer to the :ref:`Association Generator <design-generator>`
    documentation for a full description.
    """
    associations = []
    if isinstance(version_id, bool):
        version_id = make_timestamp()
    process_queue = ProcessQueueSorted(
        [ProcessList(items=pool, rules=[rule for _, rule in rules.items()])]
    )

    logger.debug("Initial process queue: %s", process_queue)
    for process_list in process_queue:
        logger.debug("** Working process list: %s", process_list)
        time_start = timer()
        total_mod_existing = 0
        total_new = 0
        total_reprocess = 0
        with Bar(
            "Processing items",
            log_level=logger.getEffectiveLevel(),
            max=len(process_list.items),
        ) as bar:
            for item in process_list.items:
                item = PoolRow(item)

                existing_asns, new_asns, to_process = generate_from_item(
                    item, version_id, associations, rules, process_list
                )
                total_mod_existing += len(existing_asns)
                total_new += len(new_asns)
                associations.extend(new_asns)

                # If working on a process list EXISTING
                # remove any new `to_process` that is
                # also EXISTING. Prevent infinite loops.
                to_process_modified = []
                for next_list in to_process:
                    next_list = workover_filter(next_list, process_list.work_over)
                    if next_list:
                        to_process_modified.append(next_list)
                process_queue.extend(to_process_modified)
                total_reprocess += len(to_process_modified)
                bar.next()

        logger.debug(
            "Existing associations modified: %d New associations created: %d",
            total_mod_existing,
            total_new,
        )
        logger.debug("New process lists: %d", total_reprocess)
        logger.debug("Updated process queue: %s", process_queue)
        logger.debug("# associations: %d", len(associations))
        logger.debug("Seconds to process: %.2f", timer() - time_start)

    # Finalize found associations
    logger.debug("# associations before finalization: %d", len(associations))
    finalized_asns = associations
    if finalize:
        logger.debug("Performing association finalization.")
        try:
            finalized_asns = rules.callback.reduce("finalize", associations)
        except KeyError as exception:
            logger.debug("Finalization failed for reason: %s", exception)

    return finalized_asns


def generate_from_item(item, version_id, associations, rules, process_list):
    """Either match or generate a new association

    Parameters
    ----------
    item : dict
        The item to match to existing associations
        or generate new associations from

    version_id : str or None
        Version id to use with association creation.
        If None, no versioning is used.

    associations : [association, ...]
        List of already existing associations.
        If the item matches any of these, it will be added
        to them.

    rules : AssociationRegistry or None
        List of rules to create new associations

    process_list : ProcessList
        The `ProcessList` from which the current item belongs to.

    Returns
    -------
    (associations, process_list): 3-tuple where
        existing_asns : [association,...]
            List of existing associations item belongs to.
            Empty if none match
        new_asns : [association,...]
            List of new associations item creates. Empty if none match
        process_list : [ProcessList, ...]
            List of process events.
    """

    # Setup the rules allowed to be examined.
    if process_list.rules is None or len(process_list.rules) == 0:
        allowed_rules = list(rules.values())
    else:
        allowed_rules = process_list.rules

    # Check membership in existing associations.
    existing_asns = []
    reprocess_list = []
    if process_list.work_over in (
        ListCategory.BOTH,
        ListCategory.EXISTING,
        ListCategory.NONSCIENCE,
    ):
        associations = [asn for asn in associations if type(asn) in allowed_rules]
        existing_asns, reprocess_list = match_item(item, associations)

    # Now see if this item will create new associations.
    # By default, a item will not be allowed to create
    # an association based on rules of existing associations.
    reprocess = []
    new_asns = []
    if (
        process_list.work_over
        in (
            ListCategory.BOTH,
            ListCategory.RULES,
        )
        and rules is not None
    ):
        ignore_asns = {type(asn) for asn in existing_asns}
        new_asns, reprocess = rules.match(
            item,
            version_id=version_id,
            allow=allowed_rules,
            ignore=ignore_asns,
        )
    reprocess_list.extend(reprocess)

    return existing_asns, new_asns, reprocess_list


def match_item(item, associations):
    """Match item to a list of associations

    Parameters
    ----------
    item : dict
        The item to match to the associations.

    associations : [association, ...]
        List of already existing associations.
        If the item matches any of these, it will be added
        to them.

    Returns
    -------
    (associations, process_list): 2-tuple where
        associations : [association,...]
            List of associations item belongs to. Empty if none match
        process_list : [ProcessList, ...]
            List of process events.
    """
    item_associations = []
    process_list = []
    for asn in associations:
        if asn in item_associations:
            continue
        matches, reprocess = asn.add(item)
        process_list.extend(reprocess)
        if matches:
            item_associations.append(asn)
    return item_associations, process_list



import logging

import numpy as np
from drizzle import cdrizzle, util

from . import resample_utils

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class GWCSDrizzle:
    """
    Combine images using the drizzle algorithm
    """

    def __init__(
        self,
        product,
        outwcs=None,
        wt_scl=None,
        pixfrac=1.0,
        kernel="square",
        fillval="INDEF",
    ):
        """
        Create a new Drizzle output object and set the drizzle parameters.

        Parameters
        ----------

        product : DataModel
            A data model containing results from a previous run. The three
            extensions SCI, WHT, and CTX contain the combined image, total counts
            and image id bitmap, respectively. The WCS of the combined image is
            also read from the SCI extension.

        outwcs : `gwcs.WCS`
            The world coordinate system (WCS) of the resampled image.  If not
            provided, the WCS is taken from product.

        wt_scl : str, optional
            How each input image should be scaled. The choices are `exptime`,
            which scales each image by its exposure time, or `expsq`, which scales
            each image by the exposure time squared.  If not set, then each
            input image is scaled by its own weight map.

        pixfrac : float, optional
            The fraction of a pixel that the pixel flux is confined to. The
            default value of 1 has the pixel flux evenly spread across the image.
            A value of 0.5 confines it to half a pixel in the linear dimension,
            so the flux is confined to a quarter of the pixel area when the square
            kernel is used.

        kernel : str, optional
            The name of the kernel used to combine the inputs. The choice of
            kernel controls the distribution of flux over the kernel. The kernel
            names are: "square", "gaussian", "point", "tophat", "turbo", "lanczos2",
            and "lanczos3". The square kernel is the default.

        fillval : str, optional
            The value a pixel is set to in the output if the input image does
            not overlap it. The default value of INDEF does not set a value.
        """

        # Initialize the object fields
        self._product = product

        self.outexptime = 0.0
        self.uniqid = 0

        self.wt_scl = "" if wt_scl is None else wt_scl
        self.kernel = kernel
        self.fillval = fillval
        self.pixfrac = pixfrac

        self.sciext = "SCI"
        self.whtext = "WHT"
        self.conext = "CON"

        out_units = "cps"

        self.outexptime = getattr(product.meta.resample, "product_exposure_time", 0.0)

        self.outsci = product.data
        self.outwcs = outwcs or product.meta.wcs
        self.outwht = product.weight
        self.outcon = product.context.astype(np.int32)

        if self.outcon.ndim == 2:
            self.outcon = np.reshape(
                self.outcon, (1, self.outcon.shape[0], self.outcon.shape[1])
            )
        elif self.outcon.ndim != 3:
            raise ValueError(
                "Drizzle context image has wrong dimensions: \
                {}".format(
                    product
                )
            )

        # Check field values
        if not self.outwcs:
            raise ValueError("Either an existing file or wcs must be supplied")

        if out_units == "counts":
            np.divide(self.outsci, self.outexptime, self.outsci)
        elif out_units != "cps":
            raise ValueError(f"Illegal value for out_units: {out_units}")

    # Since the context array is dynamic, it must be re-assigned
    # back to the product's `con` attribute.
    @property
    def outcon(self):
        """Return the context array"""
        return self._product.context

    @outcon.setter
    def outcon(self, value):
        """Set new context array"""
        self._product.context = value

    def add_image(
        self,
        insci,
        inwcs,
        inwht=None,
        xmin=0,
        xmax=0,
        ymin=0,
        ymax=0,
        expin=1.0,
        in_units="cps",
        wt_scl=1.0,
    ):
        """
        Combine an input image with the output drizzled image.

        Instead of reading the parameters from a fits file, you can set
        them by calling this lower level method. `Add_fits_file` calls
        this method after doing its setup.

        Parameters
        ----------

        insci : array
            A 2d numpy array containing the input image to be drizzled.
            it is an error to not supply an image.

        inwcs : wcs
            The world coordinate system of the input image. This is
            used to convert the pixels to the output coordinate system.

        inwht : array, optional
            A 2d numpy array containing the pixel by pixel weighting.
            Must have the same dimensions as insci. If none is supplied,
            the weighting is set to one.

        xmin : float, optional
            This and the following three parameters set a bounding rectangle
            on the output image. Only pixels on the output image inside this
            rectangle will have their flux updated. Xmin sets the minimum value
            of the x dimension. The x dimension is the dimension that varies
            quickest on the image. If the value is zero or less, no minimum will
            be set in the x dimension. All four parameters are zero based,
            counting starts at zero.

        xmax : float, optional
            Sets the maximum value of the x dimension on the bounding box
            of the output image. If the value is zero or less, no maximum will
            be set in the x dimension.

        ymin : float, optional
            Sets the minimum value in the y dimension on the bounding box. The
            y dimension varies less rapidly than the x and represents the line
            index on the output image. If the value is zero or less, no minimum
            will be set in the y dimension.

        ymax : float, optional
            Sets the maximum value in the y dimension. If the value is zero or
            less, no maximum will be set in the y dimension.

        expin : float, optional
            The exposure time of the input image, a positive number. The
            exposure time is used to scale the image if the units are counts and
            to scale the image weighting if the drizzle was initialized with
            wt_scl equal to "exptime" or "expsq."

        in_units : str, optional
            The units of the input image. The units can either be "counts"
            or "cps" (counts per second.) If the value is counts, before using
            the input image it is scaled by dividing it by the exposure time.

        wt_scl : float, optional
            If drizzle was initialized with wt_scl left blank, this value will
            set a scaling factor for the pixel weighting. If drizzle was
            initialized with wt_scl set to "exptime" or "expsq", the exposure time
            will be used to set the weight scaling and the value of this parameter
            will be ignored.
        """
        if self.wt_scl == "exptime":
            wt_scl = expin
        elif self.wt_scl == "expsq":
            wt_scl = expin * expin

        wt_scl = 1.0  # hard-coded for JWST count-rate data
        self.increment_id()

        dodrizzle(
            insci,
            inwcs,
            inwht,
            self.outwcs,
            self.outsci,
            self.outwht,
            self.outcon,
            expin,
            in_units,
            wt_scl,
            uniqid=self.uniqid,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            pixfrac=self.pixfrac,
            kernel=self.kernel,
            fillval=self.fillval,
        )

    def increment_id(self):
        """
        Increment the id count and add a plane to the context image if needed

        Drizzle tracks which input images contribute to the output image
        by setting a bit in the corresponding pixel in the context image.
        The uniqid indicates which bit. So it must be incremented each time
        a new image is added. Each plane in the context image can hold 32 bits,
        so after each 32 images, a new plane is added to the context.
        """

        # Compute what plane of the context image this input would
        # correspond to:
        planeid = int(self.uniqid / 32)

        # Add a new plane to the context image if planeid overflows

        if self.outcon.shape[0] == planeid:
            plane = np.zeros_like(self.outcon[0])
            plane = plane.reshape((1, plane.shape[0], plane.shape[1]))
            self.outcon = np.concatenate((self.outcon, plane))

        # Increment the id
        self.uniqid += 1


def dodrizzle(
    insci,
    input_wcs,
    inwht,
    output_wcs,
    outsci,
    outwht,
    outcon,
    expin,
    in_units,
    wt_scl,
    uniqid=1,
    xmin=0,
    xmax=0,
    ymin=0,
    ymax=0,
    pixfrac=1.0,
    kernel="square",
    fillval="INDEF",
):
    """
    Low level routine for performing 'drizzle' operation on one image.

    Parameters
    ----------

    insci : 2d array
        A 2d numpy array containing the input image to be drizzled.

    input_wcs : gwcs.WCS object
        The world coordinate system of the input image.

    inwht : 2d array
        A 2d numpy array containing the pixel by pixel weighting.
        Must have the same dimensions as insci. If none is supplied,
        the weighting is set to one.

    output_wcs : gwcs.WCS object
        The world coordinate system of the output image.

    outsci : 2d array
        A 2d numpy array containing the output image produced by
        drizzling. On the first call it should be set to zero.
        Subsequent calls it will hold the intermediate results

    outwht : 2d array
        A 2d numpy array containing the output counts. On the first
        call it should be set to zero. On subsequent calls it will
        hold the intermediate results.

    outcon : 2d or 3d array, optional
        A 2d or 3d numpy array holding a bitmap of which image was an input
        for each output pixel. Should be integer zero on first call.
        Subsequent calls hold intermediate results.

    expin : float
        The exposure time of the input image, a positive number. The
        exposure time is used to scale the image if the units are counts.

    in_units : str
        The units of the input image. The units can either be "counts"
        or "cps" (counts per second.)

    wt_scl : float
        A scaling factor applied to the pixel by pixel weighting.

    uniqid : int, optional
        The id number of the input image. Should be one the first time
        this function is called and incremented by one on each subsequent
        call.

    xmin : float, optional
        This and the following three parameters set a bounding rectangle
        on the input image. Only pixels on the input image inside this
        rectangle will have their flux added to the output image. Xmin
        sets the minimum value of the x dimension. The x dimension is the
        dimension that varies quickest on the image. If the value is zero,
        no minimum will be set in the x dimension. All four parameters are
        zero based, counting starts at zero.

    xmax : float, optional
        Sets the maximum value of the x dimension on the bounding box
        of the input image. If the value is zero, no maximum will
        be set in the x dimension, the full x dimension of the output
        image is the bounding box.

    ymin : float, optional
        Sets the minimum value in the y dimension on the bounding box. The
        y dimension varies less rapidly than the x and represents the line
        index on the input image. If the value is zero, no minimum  will be
        set in the y dimension.

    ymax : float, optional
        Sets the maximum value in the y dimension. If the value is zero, no
        maximum will be set in the y dimension,  the full x dimension
        of the output image is the bounding box.

    pixfrac : float, optional
        The fraction of a pixel that the pixel flux is confined to. The
        default value of 1 has the pixel flux evenly spread across the image.
        A value of 0.5 confines it to half a pixel in the linear dimension,
        so the flux is confined to a quarter of the pixel area when the square
        kernel is used.

    kernel: str, optional
        The name of the kernel used to combine the input. The choice of
        kernel controls the distribution of flux over the kernel. The kernel
        names are: "square", "gaussian", "point", "tophat", "turbo", "lanczos2",
        and "lanczos3". The square kernel is the default.

    fillval: str, optional
        The value a pixel is set to in the output if the input image does
        not overlap it. The default value of INDEF does not set a value.

    Returns
    -------
    A tuple with three values: a version string, the number of pixels
    on the input image that do not overlap the output image, and the
    number of complete lines on the input image that do not overlap the
    output input image.

    """

    # Insure that the fillval parameter gets properly interpreted for use with tdriz
    if util.is_blank(str(fillval)):
        fillval = "INDEF"
    else:
        fillval = str(fillval)

    if in_units == "cps":
        expscale = 1.0
    else:
        expscale = expin

    if insci.dtype > np.float32:
        insci = insci.astype(np.float32)

    # Add input weight image if it was not passed in
    if inwht is None:
        inwht = np.ones_like(insci)

    if xmax is None or xmax == xmin:
        xmax = insci.shape[1]
    if ymax is None or ymax == ymin:
        ymax = insci.shape[0]

    # Compute what plane of the context image this input would
    # correspond to:
    planeid = int((uniqid - 1) / 32)

    # Check if the context image has this many planes
    if outcon.ndim == 3:
        nplanes = outcon.shape[0]
    elif outcon.ndim == 2:
        nplanes = 1
    else:
        nplanes = 0

    if nplanes <= planeid:
        raise IndexError("Not enough planes in drizzle context image")

    # Alias context image to the requested plane if 3d
    if outcon.ndim == 3:
        outcon = outcon[planeid]

    # Compute the mapping between the input and output pixel coordinates
    # for use in drizzle.cdrizzle.tdriz
    pixmap = resample_utils.calc_gwcs_pixmap(input_wcs, output_wcs, insci.shape)
    # inwht[np.isnan(pixmap[:,:,0])] = 0.

    log.debug(f"Pixmap shape: {pixmap[:,:,0].shape}")
    log.debug(f"Input Sci shape: {insci.shape}")
    log.debug(f"Output Sci shape: {outsci.shape}")

    # Call 'drizzle' to perform image combination
    log.info(f"Drizzling {insci.shape} --> {outsci.shape}")

    _vers, nmiss, nskip = cdrizzle.tdriz(
        insci.astype(np.float32),
        inwht.astype(np.float32),
        pixmap,
        outsci,
        outwht,
        outcon,
        uniqid=uniqid,
        xmin=xmin,
        xmax=xmax,
        ymin=ymin,
        ymax=ymax,
        pixfrac=pixfrac,
        kernel=kernel,
        in_units=in_units,
        expscale=expscale,
        wtscale=wt_scl,
        fillstr=fillval,
    )
    log.info(
        f"Results from cdrizzle.tdriz(): \
            '_vers'={_vers}, 'nmiss'={nmiss}, 'nskip'={nskip}"
    )
    return _vers, nmiss, nskip



""" Main module for associations """

import json
import logging
import re
from collections.abc import MutableMapping
from copy import deepcopy
from datetime import datetime

import jsonschema
from stpipe.format_template import FormatTemplate

from . import __version__
from .exceptions import AssociationNotValidError
from .lib.constraint import Constraint, meets_conditions
from .lib.ioregistry import IORegistry

__all__ = ["Association"]


# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Timestamp template
_TIMESTAMP_TEMPLATE = "%Y%m%dt%H%M%S"


class Association(MutableMapping):
    """Association Base Class

    Parameters
    ----------
    version_id : str or None
        Version ID to use in the name of this association.
        If None, nothing is added.

    Raises
    ------
    AssociationError
        If an item doesn't match.

    Attributes
    ----------
    instance : dict-like
        The instance is the association data structure.
        See `data` below

    meta : dict
        Information about the association.

    data : dict
        The association. The format of this data structure
        is determined by the individual associations and, if
        defined, validated against their specified schema.

    schema_file : str
        The name of the output schema that an association
        must adhere to.
    """

    registry = None
    """Registry this rule has been placed in."""

    DEFAULT_FORCE_UNIQUE = False
    """Default whether to force constraints to use unique values."""

    DEFAULT_REQUIRE_CONSTRAINT = True
    """Default require that the constraint exists or otherwise
    can be explicitly checked.
    """

    DEFAULT_EVALUATE = False
    """Default do not evaluate input values"""

    GLOBAL_CONSTRAINT = None
    """Global constraints"""

    INVALID_VALUES = None
    """Attribute values that indicate the
    attribute is not specified.
    """

    ioregistry = IORegistry()
    """The association IO registry"""

    def __init__(
        self,
        version_id=None,
        target=None,
    ):
        self.data = {}
        self.run_init_hook = True
        self.meta = {}

        self.version_id = version_id
        self.target = target
        self.data.update(
            {
                "asn_type": "None",
                "asn_rule": self.asn_rule,
                "version_id": self.version_id,
                "code_version": __version__,
                "target": self.target,
            }
        )

        # Setup constraints
        # These may be predefined by a rule.
        try:
            constraints = self.constraints
        except AttributeError:
            constraints = Constraint()
        if self.GLOBAL_CONSTRAINT is not None:
            constraints.append(self.GLOBAL_CONSTRAINT.copy())
        self.constraints = constraints

    @classmethod
    def create(cls, item, version_id=None):
        """Create association if item belongs

        Parameters
        ----------
        item : dict
            The item to initialize the association with.

        version_id : str or None
            Version ID to use in the name of this association.
            If None, nothing is added.

        Returns
        -------
        (association, reprocess_list)
            2-tuple consisting of:
                - association or None: The association or, if the item does not
                  match this rule, None
                - [ProcessList[, ...]]: List of items to process again.
        """
        asn = cls(version_id=version_id)

        matches, reprocess = asn.add(item)
        if not matches:
            return None, reprocess
        return asn, reprocess

    @property
    def asn_name(self):
        """Suggest filename for the association"""
        return "unnamed_association"

    @classmethod
    def _asn_rule(cls):
        return cls.__name__

    @property
    def asn_rule(self):
        """Name of the rule"""
        return self._asn_rule()

    @classmethod
    def validate(cls, asn):
        """Validate an association against this rule

        Parameters
        ----------
        asn : Association or association-like
            The association structure to examine

        Returns
        -------
        valid : bool
            True if valid. Otherwise the `AssociationNotValidError` is raised

        Raises
        ------
        AssociationNotValidError
            If there is some reason validation failed.

        Notes
        -----
        The base method checks against the rule class' schema
        If the rule class does not define a schema, a warning is issued
        but the routine will return True.
        """
        if not hasattr(cls, "schema_file"):
            logger.warning(f"Cannot validate: {cls} has no schema. Presuming OK.")
            return True

        if isinstance(asn, cls):
            asn_data = asn.data
        else:
            asn_data = asn

        with open(cls.schema_file) as schema_file:
            asn_schema = json.load(schema_file)

        try:
            jsonschema.validate(asn_data, asn_schema)
        except (AttributeError, jsonschema.ValidationError) as err:
            logger.debug("Validation failed:")
            logger.debug("%s", err)
            raise AssociationNotValidError("Validation failed") from err
        return True

    def dump(self, format="json", **kwargs):
        """Serialize the association

        Parameters
        ----------
        format : str
            The format to use to dump the association into.

        kwargs : dict
            List of arguments to pass to the registered
            routines for the current association type.

        Returns
        -------
        (name, serialized):
            Tuple where the first item is the suggested
            base name for the file.
            Second item is the serialization.

        Raises
        ------
        AssociationError
            If the operation cannot be done

        AssociationNotValidError
            If the given association does not validate.
        """
        if self.is_valid:
            return self.ioregistry[format].dump(self, **kwargs)

        raise AssociationNotValidError(f"Association {self} is not valid")

    @classmethod
    def load(cls, serialized, format=None, validate=True, **kwargs):
        """Marshall a previously serialized association

        Parameters
        ----------
        serialized : object
            The serialized form of the association.

        format : str or None
            The format to force. If None, try all available.

        validate : bool
            Validate against the class' defined schema, if any.

        kwargs : dict
            Other arguments to pass to the `load` method

        Returns
        -------
        association : Association
            The association.

        Raises
        ------
        AssociationNotValidError
            Cannot create or validate the association.

        Notes
        -----
        The `serialized` object can be in any format
        supported by the registered I/O routines. For example, for
        `json` and `yaml` formats, the input can be either a string or
        a file object containing the string.
        """
        if format is None:
            formats = [
                format_func for format_name, format_func in cls.ioregistry.items()
            ]
        else:
            formats = [cls.ioregistry[format]]

        for format_func in formats:
            try:
                asn = format_func.load(cls, serialized, **kwargs)
            except AssociationNotValidError:
                continue
            else:
                break
        else:
            raise AssociationNotValidError(
                f'Cannot translate "{serialized}" to an association'
            )

        # Validate
        if validate:
            cls.validate(asn)

        return asn

    @property
    def is_valid(self):
        """Check if association is valid"""
        try:
            self.__class__.validate(self)
        except AssociationNotValidError:
            return False
        return True

    def add(self, item, check_constraints=True):
        """Add the item to the association

        Parameters
        ----------
        item : dict
            The item to add.

        check_constraints : bool
            If True, see if the item should belong to this association.
            If False, just add it.

        Returns
        -------
        (match, reprocess_list)
            2-tuple consisting of:
                - bool : True if match
                - [ProcessList[, ...]]: List of items to process again.
        """
        if self.is_item_member(item):
            return True, []

        match = not check_constraints
        if check_constraints:
            match, reprocess = self.check_and_set_constraints(item)

        if match:
            if self.run_init_hook:
                self._init_hook(item)
                self.run_init_hook = False
            self._add(item)

        # If a constraint `force_match` exists, set the `match`
        # result to the value of the constraint.
        try:
            force_match = self.constraints["force_match"].value
        except (KeyError, TypeError):
            pass
        else:
            if force_match is not None:
                match = force_match

        return match, reprocess

    def check_and_set_constraints(self, item):
        """Check whether the given dictionaries match parameters for
        for this association

        Parameters
        ----------
        item : dict
            The parameters to check/set for this association.
            This can be a list of dictionaries.

        Returns
        -------
        (match, reprocess)
            2-tuple consisting of:
                - bool : Did constraint match?
                - [ProcessItem[, ...]]: List of items to process again.

        """
        cached_constraints = deepcopy(self.constraints)
        match, reprocess = cached_constraints.check_and_set(item)
        if match:
            self.constraints = cached_constraints

        # Set the association type for all reprocessed items.
        for process_list in reprocess:
            if process_list.rules is None:
                process_list.rules = [type(self)]

        return match, reprocess

    def match_constraint(self, item, constraint, conditions):
        """Generic constraint checking

        Parameters
        ----------
        item : dict
            The item to retrieve the values from

        constraint : str
            The name of the constraint

        conditions : dict
            The conditions structure

        Returns
        -------
        (matches, reprocess_list)
            2-tuple consisting of:
                - bool : True if the all constraints are satisfied
                - [ProcessList[, ...]]: List of items to process again.
        """
        reprocess = []
        evaled_str = conditions["inputs"](item)
        if conditions["value"] is not None:
            if not meets_conditions(evaled_str, conditions["value"]):
                return False, reprocess

        # At this point, the constraint has passed.
        # Fix the conditions.
        escaped_value = re.escape(evaled_str)
        conditions["found_values"].add(escaped_value)
        if conditions["value"] is None or conditions.get(
            "force_unique", self.DEFAULT_FORCE_UNIQUE
        ):
            conditions["value"] = escaped_value
            conditions["force_unique"] = False

        # That's all folks
        return True, reprocess

    def finalize(self):
        """Finalize association

        Finalize or close-off this association. Perform validations,
        modifications, etc. to ensure that the association is
        complete.

        Returns
        -------
        associations : [association[, ...]] or None
            List of fully-qualified associations that this association
            represents.
            `None` if a complete association cannot be produced.

        """
        if self.is_valid:
            return [self]

        return None

    def is_item_member(self, item):
        """Check if item is already a member of this association

        Parameters
        ----------
        item : dict
            The item to add.

        Returns
        -------
        is_item_member : bool
            True if item is a member.
        """
        raise NotImplementedError(
            "Association.is_item_member must be implemented by a specific association"
            " rule."
        )

    def _init_hook(self, item):
        """Post-check and pre-item-adding initialization."""
        pass

    def _add(self, item):
        """Add a item, association-specific"""
        raise NotImplementedError(
            "Association._add must be implemented by a specific association rule."
        )

    def _add_items(self, items, **kwargs):
        """Force adding items to the association

        Parameters
        ----------
        items : [object[, ...]]
            A list of items to make members of the association.

        Notes
        -----
        This is a low-level shortcut into adding members, such as file names,
        to an association. All defined shortcuts and other initializations are
        by-passed, resulting in a potentially unusable association.
        """
        try:
            self["members"].update(items)
        except KeyError:
            self["members"] = items

    # #################################################
    # Methods required for implementing MutableMapping
    # #################################################
    def __getitem__(self, key):
        return self.data[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.data[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.data[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __keytransform__(self, key):
        return key

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def values(self):
        return self.data.values()


# #########
# Utilities
# #########
def finalize(asns):
    """Finalize associations by calling their `finalize_hook` method

    Notes
    -----
    This is a functioning example of a finalize callback, and can be used
    as the generic callback. Suggested usage is as follows:

    .. code-block:: python

       from romancal.associations.association import finalize as generic_finalize
       RegistryMarker.callback('finalize')(generic_finalize)
    """
    finalized_asns = list(
        filter(lambda asn: asn is not None, map(lambda asn: asn.finalize(), asns))
    )
    return finalized_asns


def make_timestamp():
    """Generate a timestamp based on runtime"""
    timestamp = datetime.utcnow().strftime(_TIMESTAMP_TEMPLATE)
    return timestamp


# Define default product name filling
format_product = FormatTemplate()



"""Association Definitions: DMS Level2b product associations
"""

import logging

from romancal.associations.lib.constraint import Constraint
from romancal.associations.lib.rules_elpp_base import *
from romancal.associations.registry import RegistryMarker

__all__ = [
    "Asn_Lv2FOV",
    "Asn_Lv2Image",
    "Asn_Lv2GBTDSPass",
    "Asn_Lv2GBTDSFull",
    "AsnMixin_Lv2Image",
    "AsnMinxin_Lv2FOV",
]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# flake8: noqa: F403, F405
# --------------------------------
# Start of the User-level rules
# --------------------------------
@RegistryMarker.rule
class Asn_Lv2FOV(AsnMixin_Lv2FOV, DMS_ELPP_Base):
    """Level2b Non-TSO Science Image Association

    Characteristics:
        - Association type: ``FOV``
        - Pipeline: ``mosaic``
        - Image-based science exposures
        - Science exposures for all 18 detectors
    """

    def __init__(self, *args, **kwargs):
        # Setup constraints
        self.constraints = Constraint(
            [
                Constraint_Base(),
                Constraint_Target(),
                Constraint_Filename(),
            ]
        )

        # Now check and continue initialization.
        super().__init__(*args, **kwargs)


@RegistryMarker.rule
class Asn_Lv2Image(AsnMixin_Lv2Image, DMS_ELPP_Base):
    """Level2b Non-TSO Science Image Association

    Characteristics:
        - Association type: ``image``
        - Pipeline: ``ELPP``
        - Image-based science exposures
        - Single science exposure
    """

    def __init__(self, *args, **kwargs):
        # Setup constraints
        self.constraints = Constraint(
            [
                Constraint_Base(),
                Constraint_Target(),
                Constraint(
                    [
                        Constraint_Expos(),
                    ],
                    reduce=Constraint.any,
                ),
                Constraint_Optical_Path(),
                Constraint_Sequence(),
                Constraint_Pass(),
                Constraint_Tile(),
            ]
        )

        # Now check and continue initialization.
        super().__init__(*args, **kwargs)


@RegistryMarker.rule
class Asn_Lv2GBTDSPass(AsnMixin_Lv2GBTDSpass, DMS_ELPP_Base):
    """Level2 GBTDS Pass Science Image Association

    Characteristics:
        - Association type: ``image``
        - Pipeline: ``romancal_image``
        - Image-based science exposures
        - Collect all exposures in a pass
        - Non-TSO
    """

    def __init__(self, *args, **kwargs):
        # Setup constraints
        self.constraints = Constraint(
            [
                Constraint_Base(),
                Constraint_Target(),
                Constraint_Category(),
                Constraint_Pass(),
                Constraint_Optical_Path(),
                Constraint_SubCategory(),
            ]
        )

        # Now check and continue initialization.
        super().__init__(*args, **kwargs)


@RegistryMarker.rule
class Asn_Lv2GBTDSFull(AsnMixin_Lv2GBTDSfull, DMS_ELPP_Base):
    """Level2 GBTDS Full Science Image Association

    Characteristics:
        - Association type: ``image``
        - Pipeline: ``romancal_image``
        - Image-based science exposures
        - Collect all exposures in a season
        - Non-TSO
    """

    def __init__(self, *args, **kwargs):
        # Setup constraints
        self.constraints = Constraint(
            [
                Constraint_Base(),
                Constraint_Target(),
                Constraint_Category(),
                Constraint_Optical_Path(),
                Constraint_SubCategory(),
            ]
        )

        # Now check and continue initialization.
        super().__init__(*args, **kwargs)

    def get_exposure_type(self, item, default=None):
        """overrides super method to return `item["exp_type"]`"""

        return item["exp_type"]



"""
Create a source catalog for tweakreg
"""

import logging

import astropy.units as u
import numpy as np
from asdf import AsdfFile
from astropy.stats import SigmaClip
from astropy.table import Table
from photutils.background import (
    Background2D,
    MeanBackground,
    MedianBackground,
    ModeEstimatorBackground,
)
from photutils.detection import DAOStarFinder
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils
from roman_datamodels.dqflags import pixel

from romancal.lib import psf
from romancal.stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["SourceDetectionStep"]


class SourceDetectionStep(RomanStep):
    """
    SourceDetectionStep: Detect point-like sources in image to create a catalog
    for alignment in tweakreg.
    """

    class_alias = "source_detection"

    spec = """
        kernel_fwhm = float(default=None)  # DAOStarFinder:Size of Gaussian kernel,
        # in pixels.
        sharplo = float(default=0.)  # DAOStarFinder: Lower bound for sharpness.
        # Typical values of sharpness range from 0 (flat) to 1 (delta function).
        sharphi = float(default=1.0)  # DAOStarFinder: Upper bound for sharpness.
        roundlo = float(default=-1.0)  # DAOStarFinder: Lower bound for roundness.
        # A circular source will have a zero roundness. A source extended in x or
        # y will have a negative or positive roundness, respectively.
        roundhi = float(default=1.0)  # DAOStarFinder: Upper bound for roundness.
        peakmax = float(default=100000.0)  # Upper limit on brightest pixel in sources.
        max_sources = float(default=None)  # Max number of sources, choosing brightest.
        scalar_threshold = float(default=None) # Detection threshold, to
        # be used for entire image. Assumed to be in same units as data, and is
        # an absolute threshold over background.
        calc_threshold = boolean(default=True) # Calculate a single absolute
        # detection threshold from image based on background.
        snr_threshold = float(default=3.0)  # if calc_threshold_img,
        # the SNR for the threshold image.
        bkg_estimator = string(default='median')  # if calc_threshold_img,
        # choice of mean, median, or mode.
        bkg_boxsize = integer(default=9)  # if calc_threshold_img,
        # size of box in pixels for 2D background.
        bkg_sigma = float(default=2.0) # if calc_threshold_img,
        # n sigma for sigma clipping bkgrnd.
        bkg_filter_size = integer(default=3) # if calc_threshold_img,
        # size of Gaussian kernel for background.
        save_catalogs = boolean(default=False) # Save source catalog to file?
        # Will overwrite an existing catalog of the same name.
        output_cat_filetype = option('asdf', 'ecsv', default='asdf') # Used if
        #save_catalogs=True - file type of output catalog.
        fit_psf = boolean(default=True)  # fit source PSFs for accurate astrometry
        min_separation = integer(default=11)  # don't find multiple sources closer
        # than this number of pixels
    """

    def process(self, input):
        if isinstance(input, rdm.DataModel):
            input_model = input
        else:
            input_model = rdm.open(input)

        if input_model.meta.exposure.type != "WFI_IMAGE":
            # Check to see if attempt to find sources in non-Image data
            log.info("Skipping source detection for spectral exposure.")
            input_model.meta.cal_step.source_detection = "SKIPPED"
            return input_model

        # remove units from data in this step.
        # DAOStarFinder requires unitless input
        if hasattr(input_model.data, "unit"):
            self.data = input_model.data.value
        else:
            self.data = input_model.data

        # mask DO_NOT_USE pixels

        self.coverage_mask = ((pixel.DO_NOT_USE) & input_model.dq).astype(bool)

        filt = input_model.meta.instrument["optical_element"]

        # if a pre-determined threshold value for detection for the whole
        # image is provided, use this
        if self.scalar_threshold is not None:
            threshold = float(self.scalar_threshold)
            log.info(f"Using a detection threshold of {threshold}.")

        # otherwise, if specified, calculate a scalar threshold from the
        # image by calculating a 2D background image, using this to create
        # a 2d threshold image, and using the median of the 2d threshold
        # image as the scalar detection threshold for the whole image
        elif self.calc_threshold:
            log.info("Determining detection threshold from image.")
            bkg = self._calc_2D_background()
            threshold_img = bkg.background + self.snr_threshold * bkg.background_rms
            threshold = np.median(threshold_img)
            log.info(f"Calculated a detection threshold of {threshold} from image.")
        else:
            threshold = np.median(self.data)

        if self.kernel_fwhm is None:
            # estimate the FWHM of the PSF using rough estimate
            # from the diffraction limit:
            central_wavelength = psf.filter_central_wavelengths[filt] * u.um
            diffraction_limit = float(1.22 * central_wavelength / (2.4 * u.m))
            assume_pixel_scale = 0.11 * u.arcsec / u.pix
            expect_psf_pix = np.sqrt(
                ((diffraction_limit * u.rad).to(u.arcsec) / assume_pixel_scale) ** 2
                + 1 * u.pix**2
            ).to_value(u.pix)
            self.kernel_fwhm = expect_psf_pix

        log.info("Detecting sources with DAOStarFinder, using entire image array.")
        daofind = DAOStarFinder(
            fwhm=self.kernel_fwhm,
            threshold=threshold,
            sharplo=self.sharplo,
            sharphi=self.sharphi,
            roundlo=self.roundlo,
            roundhi=self.roundhi,
            brightest=self.max_sources,
            peakmax=self.peakmax,
            min_separation=self.min_separation,
        )

        if self.scalar_threshold is not None:
            # if absolute threshold is provided
            sources = daofind(self.data, mask=self.coverage_mask)

        elif self.calc_threshold is not None:
            # subtract background from data if calculating abs. threshold
            sources = daofind(self.data - bkg.background, mask=self.coverage_mask)
        else:
            sources = daofind(self.data, mask=self.coverage_mask)

        # reduce table to minimal number of columns, just source ID,
        # positions, and fluxes
        columns = ["id", "xcentroid", "ycentroid", "flux"]

        if sources:
            catalog = sources[columns]
            log.info(f"Found {len(catalog)} sources.")
        else:
            # if no sources were detected, return an empty table
            self.log.warning("No sources detected, returning empty catalog.")
            catalog = Table(
                names=columns,
                dtype=(int, np.float64, np.float64, np.float64),
            )

        if self.fit_psf and len(sources):
            # refine astrometry by fitting PSF models to the detected sources
            log.info("Constructing a gridded PSF model.")
            detector = input_model.meta.instrument["detector"].replace("WFI", "SCA")
            gridded_psf_model, _ = psf.create_gridded_psf_model(
                filt=filt,
                detector=detector,
            )

            log.info(
                "Fitting a PSF model to sources for improved astrometric precision."
            )
            psf_photometry_table, photometry = psf.fit_psf_to_image_model(
                image_model=input_model,
                psf_model=gridded_psf_model,
                x_init=catalog["xcentroid"],
                y_init=catalog["ycentroid"],
                exclude_out_of_bounds=True,
            )

        catalog_as_recarray = catalog.as_array()

        # create meta.source detection section in file
        # if save_catalogs is True, this will be updated with the
        # attribute 'tweakreg_catalog_name' to point to the location
        # of the catalog on disk. If save_catalogs is false, this section
        # will be updated to contain the catalog to pass to TweakReg

        # tweakreg_catalog_name will be saved to the final output file,
        # while tweakreg_catalog is intended to be deleted by TweakRegStep
        input_model.meta["source_detection"] = maker_utils.mk_source_detection()

        # if 'save_catalogs'= True, also save the output catalog to a file
        # (format specified by output_cat_filetype) and add an attribute
        # to the file that contains the path to this file
        if self.save_catalogs:
            cat_filename = input_model.meta.filename.replace(".asdf", "")
            cat_filename += f"_tweakreg_catalog.{self.output_cat_filetype}"
            log.info(f"Saving catalog to file: {cat_filename}.")

            if self.output_cat_filetype == "asdf":
                tree = {"tweakreg_catalog": catalog_as_recarray}
                ff = AsdfFile(tree)
                ff.write_to(cat_filename)
            else:
                catalog.write(cat_filename, format="ascii.ecsv", overwrite=True)

            input_model.meta.source_detection["tweakreg_catalog_name"] = cat_filename
        else:
            if self.fit_psf:
                # PSF photometry centroid results are stored in an astropy table
                # in columns "x_fit" and "y_fit", which we'll rename for
                # compatibility with tweakreg:
                catalog = psf_photometry_table.copy()

                # rename the PSF photometry table's columns to match the
                # expectated columns in tweakreg:
                for old_name, new_name in zip(
                    ["x_fit", "y_fit"], ["xcentroid", "ycentroid"]
                ):
                    catalog.rename_column(old_name, new_name)

                input_model.meta.source_detection["tweakreg_catalog"] = catalog
            else:
                # only attach catalog to file if its being passed to the next step
                # and save_catalogs is false, since it is not in the schema
                input_model.meta.source_detection["tweakreg_catalog"] = (
                    catalog_as_recarray
                )
        input_model.meta.cal_step["source_detection"] = "COMPLETE"

        # just pass input model to next step - catalog is stored in meta
        return input_model

    def _calc_2D_background(self):
        """Calculates a 2D background image.

        Calculates the background value for the input image in boxes specified by
        self.bkg_box_size. A mean, median, or mode estimator may be used (set
        by `bkg_estimator`). The pixels in each box will be sigma clipped,
        using a sigma specified by `bkg_sigma`."""

        filter_size = (
            self.bkg_filter_size,
            self.bkg_filter_size,
        )  # square size specified
        box_size = np.asarray(self.bkg_boxsize).astype(int)  # must be integer

        if self.bkg_estimator == "median":
            bkg_estimator = MedianBackground()
        elif self.bkg_estimator == "mean":
            bkg_estimator = MeanBackground()
        elif self.bkg_estimator == "mode":
            bkg_estimator = ModeEstimatorBackground()
        else:
            raise ValueError("bkg_estimator must be one of 'mean', 'median', or 'mode'")

        sigma_clip = SigmaClip(self.bkg_sigma)

        try:
            bkg_2D = Background2D(
                self.data,
                box_size,
                filter_size=filter_size,
                coverage_mask=self.coverage_mask,
                sigma_clip=sigma_clip,
                bkg_estimator=bkg_estimator,
            )
        except ValueError:
            # use the entire unmasked array
            log.info(
                "Background could not be estimated in meshes. "
                "Using the entire unmasked array for background "
                f"estimation: bkg_boxsize={self.data.shape}."
            )

            bkg_2D = Background2D(
                self.data,
                self.data.shape,
                filter_size=filter_size,
                coverage_mask=self.coverage_mask,
                sigma_clip=sigma_clip,
                bkg_estimator=bkg_estimator,
                exclude_percentile=100.0,
            )

        return bkg_2D



""" Roman tests for flat field correction """

import pytest
import roman_datamodels as rdm

from romancal.pipeline.exposure_pipeline import ExposurePipeline

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(scope="module")
def run_elp(rtdata_module):
    rtdata = rtdata_module

    # The input data is from INS for stress testing at some point this should be generated
    # by INS every time new data is needed.

    input_data = "r0000201001001001001_0004_wfi01_uncal.asdf"
    rtdata.get_data(f"WFI/grism/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000201001001001001_0004_wfi01_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_elp):
    return run_elp.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def input_filename(run_elp):
    return run_elp.input


@pytest.fixture(scope="module")
def input_model(input_filename):
    with rdm.open(input_filename) as model:
        yield model


def test_output_is_image_model(output_model):
    # DMS414
    assert isinstance(output_model, rdm.datamodels.ImageModel)


def test_input_has_16_resultants(input_model):
    # DMS414
    assert len(input_model.meta.exposure.read_pattern) == 16


def test_output_has_16_resultants(output_model):
    # DMS414
    assert len(output_model.meta.exposure.read_pattern) == 16


@pytest.mark.parametrize(
    "step_name, status",
    (
        ("assign_wcs", "COMPLETE"),
        ("dark", "COMPLETE"),
        ("dq_init", "COMPLETE"),
        ("linearity", "COMPLETE"),
        ("ramp_fit", "COMPLETE"),
        ("saturation", "COMPLETE"),
        # skipped
        ("flat_field", "SKIPPED"),
        ("photom", "SKIPPED"),
    ),
)
def test_step_status(output_model, step_name, status):
    assert getattr(output_model.meta.cal_step, step_name) == status



from pathlib import Path

import numpy as np
import pytest
import roman_datamodels as rdm
from numpy.testing import assert_allclose
from roman_datamodels.dqflags import pixel

from romancal.assign_wcs.assign_wcs_step import AssignWcsStep
from romancal.pipeline.exposure_pipeline import ExposurePipeline

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(scope="module")
def run_elp(rtdata_module):
    rtdata = rtdata_module

    input_data = "r0000201001001001001_0001_wfi01_uncal.asdf"
    rtdata.get_data(f"WFI/grism/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000201001001001001_0001_wfi01_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/grism/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_elp):
    return run_elp.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_elp):
    return run_elp.truth


@pytest.fixture(scope="module")
def repointed_filename_and_delta(output_filename):
    delta = [10.0, 10.0]

    with rdm.open(output_filename) as model:
        # open from the file since we're going to modify the model
        del model.meta["wcs"]

        # Create new pointing for the model
        # RA & Dec are each shifted + 10 degrees, unless they are near
        # the upper limit, in which case they are shifted -10 degrees.
        if model.meta.wcsinfo.ra_ref >= 350.0:
            delta[0] *= -1.0
        if model.meta.wcsinfo.dec_ref >= 80.0:
            delta[1] *= -1.0

        model.meta.wcsinfo.ra_ref += delta[0]
        model.meta.wcsinfo.dec_ref += delta[1]

        # Create new wcs object for the new pointing
        model = AssignWcsStep.call(model)

        # save repointed model
        repointed_filename = output_filename.rsplit(".", 1)[0] + "_repoint.asdf"
        model.to_asdf(repointed_filename)

    return repointed_filename, delta


def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.parametrize(
    "step_name, status",
    (
        ("assign_wcs", "COMPLETE"),
        ("dark", "COMPLETE"),
        ("dq_init", "COMPLETE"),
        ("linearity", "COMPLETE"),
        ("ramp_fit", "COMPLETE"),
        ("saturation", "COMPLETE"),
        ("refpix", "COMPLETE"),
        # skipped
        ("flat_field", "SKIPPED"),
        ("photom", "SKIPPED"),
        ("source_detection", "SKIPPED"),
        ("tweakreg", "SKIPPED"),
    ),
)
def test_step_status(output_model, step_name, status):
    # DMS90
    # DMS278
    # also DMS129 for assign_wcs
    assert getattr(output_model.meta.cal_step, step_name) == status


def test_jump_in_uneven_ramp(output_model):
    # DMS365 jump detection detected jumps in uneven ramp
    uneven = len({len(x) for x in output_model.meta.exposure.read_pattern}) > 1
    assert uneven & np.any(output_model.dq & pixel.JUMP_DET)


@pytest.mark.parametrize("arr_name", ("dq", "err", "var_poisson", "var_rnoise"))
def test_array_exists(output_model, arr_name):
    # DMS91
    assert hasattr(output_model, arr_name)


def test_has_exposure_time(output_model):
    # DMS88 total exposure time exists
    assert "exposure_time" in output_model.meta.exposure


def test_wcs_has_bounding_box(output_model):
    # DMS93 WCS tests
    assert len(output_model.meta.wcs.bounding_box) == 2


def test_repointed_matches_truth(
    repointed_filename_and_delta, rtdata, ignore_asdf_paths
):
    # DMS90
    repointed_filename, _ = repointed_filename_and_delta

    rtdata.get_truth(f"truth/WFI/grism/{Path(repointed_filename).name}")
    rtdata.output = repointed_filename
    diff = compare_asdf(repointed_filename, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_repointed_wcs_differs(repointed_filename_and_delta, output_model):
    # DMS93 WCS tests
    repointed_filename, delta = repointed_filename_and_delta
    orig_wcs = output_model.meta.wcs
    with rdm.open(repointed_filename) as repointed_model:
        assert_allclose(
            [orig_wcs(2048, 2048)[0] + delta[0], orig_wcs(2048, 2048)[1] + delta[1]],
            repointed_model.meta.wcs(2048, 2048),
            atol=1.0,
        )



from pathlib import Path

import numpy as np
import pytest
import roman_datamodels as rdm
from gwcs.wcstools import grid_from_bounding_box
from numpy.testing import assert_allclose
from roman_datamodels.dqflags import pixel

from romancal.assign_wcs.assign_wcs_step import AssignWcsStep
from romancal.pipeline.exposure_pipeline import ExposurePipeline

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = pytest.mark.bigdata


@pytest.fixture(scope="module")
def run_elp(rtdata_module):
    rtdata = rtdata_module

    input_data = "r0000101001001001001_0001_wfi01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_0001_wfi01_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)

    # get truth file
    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_elp):
    return run_elp.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_elp):
    return run_elp.truth


@pytest.fixture(scope="module")
def repointed_filename_and_delta(output_filename):
    delta = [10.0, 10.0]

    with rdm.open(output_filename) as model:
        # open from the file since we're going to modify the model
        del model.meta["wcs"]

        # Create new pointing for the model
        # RA & Dec are each shifted + 10 degrees, unless they are near
        # the upper limit, in which case they are shifted -10 degrees.
        if model.meta.wcsinfo.ra_ref >= 350.0:
            delta[0] *= -1.0
        if model.meta.wcsinfo.dec_ref >= 80.0:
            delta[1] *= -1.0

        model.meta.wcsinfo.ra_ref += delta[0]
        model.meta.wcsinfo.dec_ref += delta[1]

        # Create new wcs object for the new pointing
        model = AssignWcsStep.call(model)

        # save repointed model
        repointed_filename = output_filename.rsplit(".", 1)[0] + "_repoint.asdf"
        model.to_asdf(repointed_filename)

    return repointed_filename, delta


@pytest.mark.soctests
def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.soctests
def test_output_is_image_model(output_model):
    # DMS280 result is an ImageModel
    assert isinstance(output_model, rdm.datamodels.ImageModel)


@pytest.mark.soctests
@pytest.mark.parametrize(
    "step_name",
    (
        "assign_wcs",
        "flat_field",
        "dark",
        "dq_init",
        "linearity",
        "ramp_fit",
        "saturation",
    ),
)
def test_steps_ran(output_model, step_name):
    # DMS86
    # also DMS129 for assign_wcs
    assert getattr(output_model.meta.cal_step, step_name) == "COMPLETE"


@pytest.mark.soctests
def test_jump_in_uneven_ramp(output_model):
    # DMS361 jump detection detected jumps in uneven ramp
    uneven = len({len(x) for x in output_model.meta.exposure.read_pattern}) > 1
    assert uneven & np.any(output_model.dq & pixel.JUMP_DET)


@pytest.mark.soctests
def test_has_a_wcs(output_model):
    # DMS129
    assert output_model.meta.wcs is not None


@pytest.mark.soctests
def test_wcs_has_distortion_information(output_model):
    # DMS129
    assert "v2v3" in output_model.meta.wcs.available_frames


@pytest.mark.soctests
def test_wcs_applies_distortion_correction(output_model):
    # DMS129
    # compare coordinates before and after distortion correction has been applied
    # 1 - get new image array based on the model
    x0, y0 = grid_from_bounding_box(output_model.meta.wcs.bounding_box)
    # 2 - apply the distortion-corrected WCS solution to new image array
    corrected_coords = output_model.meta.wcs(x0, y0)
    # 3 - apply the transformation from 'v2v3' to 'world' without distortion correction
    original_coords = output_model.meta.wcs.get_transform("v2v3", "world")(x0, y0)
    # compare both results to make sure they don't match
    # (which means the distortion correction was actually applied to the model)
    assert (corrected_coords[0] != original_coords[0]).all()
    assert (corrected_coords[1] != original_coords[1]).all()


@pytest.mark.soctests
@pytest.mark.parametrize(
    "arr_name", ("dq", "err", "var_poisson", "var_rnoise", "var_flat")
)
def test_array_exists(output_model, arr_name):
    # DMS87
    assert hasattr(output_model, arr_name)


@pytest.mark.soctests
def test_has_exposure_time(output_model):
    # DMS88 total exposure time exists
    assert "exposure_time" in output_model.meta.exposure


@pytest.mark.soctests
@pytest.mark.parametrize("meta_attribute", ("detector", "optical_element"))
def test_instrument_meta(output_model, meta_attribute):
    # DMS136 PSF tests
    assert meta_attribute in output_model.meta.instrument


@pytest.mark.soctests
def test_wcs_has_bounding_box(output_model):
    # DMS89 WCS tests
    assert len(output_model.meta.wcs.bounding_box) == 2


@pytest.mark.soctests
def test_repointed_matches_truth(
    repointed_filename_and_delta, rtdata, ignore_asdf_paths
):
    # DMS89 WCS tests
    repointed_filename, _ = repointed_filename_and_delta

    rtdata.get_truth(f"truth/WFI/image/{Path(repointed_filename).name}")
    rtdata.output = repointed_filename
    diff = compare_asdf(repointed_filename, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.soctests
def test_repointed_wcs_differs(repointed_filename_and_delta, output_model):
    # DMS89 WCS tests
    repointed_filename, delta = repointed_filename_and_delta
    orig_wcs = output_model.meta.wcs
    with rdm.open(repointed_filename) as repointed_model:
        assert_allclose(
            [orig_wcs(2048, 2048)[0] + delta[0], orig_wcs(2048, 2048)[1] + delta[1]],
            repointed_model.meta.wcs(2048, 2048),
            atol=1.0,
        )


def test_elp_input_dm(rtdata, ignore_asdf_paths):
    """Test for input roman Datamodel to exposure level pipeline"""
    input_data = "r0000101001001001001_0001_wfi01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    dm_input = rdm.open(rtdata.input)

    # Test Pipeline with input datamodel
    output = "r0000101001001001001_0001_wfi01_cal.asdf"
    rtdata.output = output
    ExposurePipeline.call(dm_input, save_results=True)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    # Ensure step completion is as expected
    with rdm.open(rtdata.output) as model:
        assert model.meta.cal_step.dq_init == "COMPLETE"
        assert model.meta.cal_step.saturation == "COMPLETE"
        assert model.meta.cal_step.linearity == "COMPLETE"
        assert model.meta.cal_step.dark == "COMPLETE"
        assert model.meta.cal_step.ramp_fit == "COMPLETE"
        assert model.meta.cal_step.assign_wcs == "COMPLETE"
        assert model.meta.cal_step.flat_field == "COMPLETE"
        assert model.meta.cal_step.photom == "COMPLETE"


def test_processing_pipeline_all_saturated(rtdata, ignore_asdf_paths):
    """Tests for fully saturated data skipping steps in the pipeline

    Note that this test mimics how the pipeline is run in OPS.
    Any changes to this test should be coordinated with OPS.
    """
    input_data = "r0000101001001001001_0001_wfi01_ALL_SATURATED_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_0001_wfi01_ALL_SATURATED_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Ensure step completion is as expected
    with rdm.open(rtdata.output) as model:
        assert model.meta.cal_step.dq_init == "COMPLETE"
        assert model.meta.cal_step.saturation == "COMPLETE"
        assert model.meta.cal_step.linearity == "SKIPPED"
        assert model.meta.cal_step.dark == "SKIPPED"
        assert model.meta.cal_step.jump == "SKIPPED"
        assert model.meta.cal_step.ramp_fit == "SKIPPED"
        assert model.meta.cal_step.assign_wcs == "SKIPPED"
        assert model.meta.cal_step.flat_field == "SKIPPED"
        assert model.meta.cal_step.photom == "SKIPPED"


def test_pipeline_suffix(rtdata, ignore_asdf_paths):
    """
    Tests passing suffix to the pipeline

    Note that this test mimics how the pipeline is run in OPS.

    Any changes to this test should be coordinated with OPS.
    """
    input_data = "r0000101001001001001_0001_wfi01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")

    output = "r0000101001001001001_0001_wfi01_star.asdf"
    rtdata.output = output

    args = [
        "roman_elp",
        rtdata.input,
        "--steps.tweakreg.skip=True",
        "--suffix=star",
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Ensure step completion is as expected
    with rdm.open(rtdata.output) as model:
        assert model.meta.cal_step.dq_init == "COMPLETE"
        assert model.meta.cal_step.saturation == "COMPLETE"
        assert model.meta.cal_step.linearity == "COMPLETE"
        assert model.meta.cal_step.dark == "COMPLETE"
        assert model.meta.cal_step.ramp_fit == "COMPLETE"
        assert model.meta.cal_step.assign_wcs == "COMPLETE"
        assert model.meta.cal_step.flat_field == "COMPLETE"
        assert model.meta.cal_step.photom == "COMPLETE"
        assert model.meta.cal_step.source_detection == "COMPLETE"
        assert model.meta.cal_step.tweakreg == "INCOMPLETE"
        assert model.meta.filename == output



""" Roman tests for source catalog creation """

import asdf
import pytest

from romancal.multiband_catalog.multiband_catalog_step import MultibandCatalogStep
from romancal.stpipe import RomanStep

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]

fieldlist = [
    "ra_centroid",  # DMS374 positions on ICRF
    "dec_centroid",  # DMS374 positions on ICRF
    "F158_aper30_flux",  # DMS399 aperture fluxes
    "F158_aper50_flux",  # DMS399 aperture fluxes
    "F158_aper70_flux",  # DMS399 aperture fluxes
    "F158_aper_total_flux",  # DMS375 fluxes
    "F158_aper_total_flux_err",  # DMS386 flux uncertainties
    "flags",  # DMS387 dq_flags
]


def test_multiband_catalog(rtdata_module):
    rtdata = rtdata_module
    inputasnfn = "L3_skycell_mbcat_asn.json"
    # note that this input association currently only has a single
    # filter in it, so this is more of an existence proof for the multiband
    # catalogs than a detailed test.  Using only a single catalog lets us
    # rely on the existing regtest files.
    outputfn = "r0099101001001001001_r274dp63x31y81_prompt_cat.asdf"
    rtdata.get_asn(f"WFI/image/{inputasnfn}")
    rtdata.output = outputfn
    rtdata.input = inputasnfn
    rtdata.get_truth(f"truth/WFI/image/{outputfn}")
    aftruth = asdf.open(f"truth/{outputfn}")
    args = [
        "romancal.step.MultibandCatalogStep",
        inputasnfn,
        "--deblend",
        "True",  # use deblending, DMS 393
    ]
    RomanStep.from_cmdline(args)
    afcat = asdf.open(outputfn)
    for field in fieldlist:
        assert field in afcat["roman"]["source_catalog"].dtype.names

    # DMS 393: multiband catalog uses both PSF-like and extend-source-like
    # kernels
    assert set(aftruth["roman"]["source_catalog"].dtype.names) == set(
        afcat["roman"]["source_catalog"].dtype.names
    )
    # weak assertion that our truth file must at least have the same
    # catalog fields as the file produced here.  Exactly matching rows
    # would require a lot of okifying things that aren't obviously
    # the same, but we can easily check that the columns match up
    # by name.

    step = MultibandCatalogStep()
    step.log.info("DMS391: successfully used multiple kernels to detect sources.")
    step.log.info("DMS393: successfully used deblending to separate blended sources.")
    step.log.info(
        "DMS399: successfully tested that catalogs contain aperture "
        "fluxes and uncertainties."
    )



"""Unit-like tests related to FluxStep"""

import astropy.units as u
import numpy as np
import pytest
from roman_datamodels import datamodels, maker_utils

from romancal.datamodels import ModelLibrary
from romancal.flux import FluxStep


@pytest.mark.parametrize(
    "attr, factor",
    [
        ("data", 1),
        ("err", 1),
        ("var_rnoise", 2),
        ("var_poisson", 2),
        ("var_flat", 2),
    ],
)
def test_attributes(flux_step, attr, factor):
    """Test that the attribute has been scaled by the right factor"""
    original, result = flux_step
    c_unit = 1.0

    # Handle difference between just a single image and a list.
    if isinstance(original, datamodels.ImageModel):
        original_library = ModelLibrary([original])
        result_library = ModelLibrary([result])
    else:
        original_library = original
        result_library = result

    assert len(original_library) == len(result_library)
    with original_library, result_library:
        for i in range(len(original_library)):
            original_model = original_library.borrow(i)
            result_model = result_library.borrow(i)

            c_mj = original_model.meta.photometry.conversion_megajanskys
            scale = (c_mj * c_unit) ** factor
            original_value = getattr(original_model, attr)
            result_value = getattr(result_model, attr)

            assert np.allclose(original_value * scale, result_value)

            original_library.shelve(original_model, i, modify=False)
            result_library.shelve(result_model, i, modify=False)


# ########
# Fixtures
# ########
@pytest.fixture(scope="module", params=["input_imagemodel", "input_modellibrary"])
def flux_step(request):
    """Execute FluxStep on given input

    Parameters
    ----------
    input : str, `roman_datamodels.datamodels.DataModel`, or `~romancal.datamodels.library.ModelLibrary`

    Returns
    -------
    original, result : DataModel or ModelLibrary, DataModel or ModelLibrary
    """
    init = request.getfixturevalue(request.param)
    if isinstance(init, ModelLibrary):
        models = []
        with init:
            for m in init:
                models.append(m.copy())
                init.shelve(m, modify=False)
        original = ModelLibrary(models)
    else:
        original = init.copy()

    # Perform step
    result = FluxStep.call(init)

    # That's all folks
    return original, result


@pytest.fixture(scope="module")
def image_model():
    """Product a basic ImageModel"""
    # Create a random image and specify a conversion
    rng = np.random.default_rng()
    shape = (10, 10)
    image_model = maker_utils.mk_datamodel(datamodels.ImageModel, shape=shape)
    image_model.data = rng.poisson(2.5, size=shape).astype(np.float32)
    image_model.var_rnoise = rng.normal(1, 0.05, size=shape).astype(np.float32)
    image_model.var_poisson = rng.poisson(1, size=shape).astype(np.float32)
    image_model.var_flat = rng.uniform(0, 1, size=shape).astype(np.float32)
    image_model.meta.photometry.conversion_megajanskys = (2.0 * u.MJy / u.sr).value

    return image_model


@pytest.fixture(scope="module")
def input_imagemodel(image_model):
    """Provide a single ImageModel"""

    # First just setup the basic model
    return image_model.copy()


@pytest.fixture(scope="module")
def input_modellibrary(image_model):
    """Provide a ModelLibrary"""
    # Create and return a ModelLibrary
    image_model1 = image_model.copy()
    image_model2 = image_model.copy()
    image_model2.meta.photometry.conversion_megajanskys = (0.5 * u.MJy / u.sr).value
    container = ModelLibrary([image_model1, image_model2])
    return container



import copy
import logging
from functools import partial

import numpy as np
from roman_datamodels.dqflags import pixel
from stcal.outlier_detection.median import MedianComputer
from stcal.outlier_detection.utils import (
    compute_weight_threshold,
    flag_crs,
    flag_resampled_crs,
    gwcs_blot,
)

from romancal.resample.resample import ResampleData
from romancal.resample.resample_utils import build_driz_weight

from . import _fileio

__all__ = ["detect_outliers"]


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def _median_with_resampling(
    input_models,
    resamp,
    maskpt,
    save_intermediate_results,
    make_output_path,
    buffer_size=None,
):
    """
    Compute median of resampled data from models in a library.

    Parameters
    ----------
    input_models : ModelLibrary
        The input datamodels.

    resamp : resample.resample.ResampleData object
        The controlling object for the resampling process.

    maskpt : float
        The weight threshold for masking out low weight pixels.

    save_intermediate_results : bool
        if True, save the drizzled models and median model to fits.

    make_output_path : function
        The functools.partial instance to pass to save_median.

    buffer_size : int
        The size of chunk in bytes that will be read into memory when computing the median.
        This parameter has no effect if the input library has its on_disk attribute
        set to False. If None or 0 the buffer size will be set to the size of one
        resampled image.
    """
    if not resamp.single:
        raise ValueError(
            "median_with_resampling should only be used for resample_many_to_many"
        )

    in_memory = not input_models._on_disk
    indices_by_group = list(input_models.group_indices.values())
    nresultants = len(indices_by_group)
    example_model = None
    median_wcs = resamp.output_wcs

    with input_models:
        for i, indices in enumerate(indices_by_group):

            drizzled_model = resamp.resample_group(input_models, indices)

            if save_intermediate_results:
                # write the drizzled model to file
                _fileio.save_drizzled(drizzled_model, make_output_path)

            if i == 0:
                input_shape = (nresultants,) + drizzled_model.data.shape
                dtype = drizzled_model.data.dtype
                computer = MedianComputer(input_shape, in_memory, buffer_size, dtype)
                example_model = drizzled_model

            weight_threshold = compute_weight_threshold(drizzled_model.weight, maskpt)
            drizzled_model.data[drizzled_model.weight < weight_threshold] = np.nan
            computer.append(drizzled_model.data, i)
            del drizzled_model

    # Perform median combination on set of drizzled mosaics
    median_data = computer.evaluate()

    if save_intermediate_results:
        # drizzled model already contains asn_id
        _fileio.save_median(
            example_model,
            median_data,
            median_wcs,
            partial(make_output_path, asn_id=None),
        )

    return median_data, median_wcs


def _median_without_resampling(
    input_models,
    maskpt,
    weight_type,
    good_bits,
    save_intermediate_results,
    make_output_path,
    buffer_size=None,
):
    """
    Compute median of data from models in a library.

    Parameters
    ----------
    input_models : ModelLibrary
        The input datamodels.

    maskpt : float
        The weight threshold for masking out low weight pixels.

    weight_type : str
        The type of weighting to use when combining images. Options are:
        'ivm' (inverse variance) or 'exptime' (exposure time).

    good_bits : int
        The bit values that are considered good when determining the
        data quality of the input.

    save_intermediate_results : bool
        if True, save the models and median model to fits.

    make_output_path : function
        The functools.partial instance to pass to save_median.

    buffer_size : int
        The size of chunk in bytes that will be read into memory when computing the median.
        This parameter has no effect if the input library has its on_disk attribute
        set to False. If None or 0 the buffer size will be set to the size of one
        input image.

    """
    in_memory = not input_models._on_disk
    nresultants = len(input_models)
    example_model = None

    with input_models:
        for i in range(len(input_models)):

            model = input_models.borrow(i)
            wht = build_driz_weight(
                model,
                # FIXME this was hard-coded to "ivm"
                weight_type="ivm",
                good_bits=good_bits,
            )

            if save_intermediate_results:
                # write the model to file
                _fileio.save_drizzled(model, make_output_path)

            if i == 0:
                input_shape = (nresultants,) + model.data.shape
                dtype = model.data.dtype
                computer = MedianComputer(input_shape, in_memory, buffer_size, dtype)
                example_model = model
                median_wcs = copy.deepcopy(model.meta.wcs)

            weight_threshold = compute_weight_threshold(wht, maskpt)

            data_copy = model.data.copy()
            data_copy[wht < weight_threshold] = np.nan
            computer.append(data_copy, i)

            input_models.shelve(model, i, modify=True)
            del data_copy, model

    # Perform median combination on set of drizzled mosaics
    median_data = computer.evaluate()

    if save_intermediate_results:
        _fileio.save_median(example_model, median_data, median_wcs, make_output_path)

    return median_data, median_wcs


def _flag_resampled_model_crs(
    image,
    median_data,
    median_wcs,
    snr1,
    snr2,
    scale1,
    scale2,
    backg,
    save_intermediate_results,
    make_output_path,
):
    blot = gwcs_blot(median_data, median_wcs, image.data.shape, image.meta.wcs, 1.0)

    # Get background level of science data if it has not been subtracted, so it
    # can be added into the level of the blotted data, which has been
    # background-subtracted
    if (
        hasattr(image.meta, "background")
        and image.meta.background.subtracted is False
        and image.meta.background.level is not None
    ):
        backg = image.meta.background.level
        log.debug(
            f"Adding background level {image.meta.background.level} to blotted image"
        )

    cr_mask = flag_resampled_crs(
        image.data, image.err, blot, snr1, snr2, scale1, scale2, backg
    )

    # update the dq flags in-place
    image.dq |= cr_mask * np.uint32(pixel.DO_NOT_USE | pixel.OUTLIER)
    log.info(f"{np.count_nonzero(cr_mask)} pixels marked as outliers")


def _flag_model_crs(image, median_data, snr):
    cr_mask = flag_crs(image.data, image.err, median_data, snr)

    # Update dq array in-place
    image.dq |= cr_mask * np.uint32(pixel.DO_NOT_USE | pixel.OUTLIER)

    log.info(f"{np.count_nonzero(cr_mask)} pixels marked as outliers")


def detect_outliers(
    library,
    weight_type,
    pixfrac,
    kernel,
    fillval,
    maskpt,
    snr1,
    snr2,
    scale1,
    scale2,
    backg,
    save_intermediate_results,
    resample_data,
    good_bits,
    in_memory,
    make_output_path,
):
    # setup ResampleData
    # call
    if resample_data:
        resamp = ResampleData(
            library,
            single=True,
            blendheaders=False,
            # FIXME prior code provided weight_type when only wht_type is understood
            # both default to 'ivm' but tests that set this to something else did
            # not change the resampling weight type. For now, disabling it to match
            # the previous behavior.
            # wht_type=weight_type
            pixfrac=pixfrac,
            kernel=kernel,
            fillval=fillval,
            good_bits=good_bits,
            in_memory=in_memory,
        )
        median_data, median_wcs = _median_with_resampling(
            library,
            resamp,
            maskpt,
            save_intermediate_results=save_intermediate_results,
            make_output_path=make_output_path,
        )
    else:
        median_data, median_wcs = _median_without_resampling(
            library,
            maskpt,
            weight_type,
            good_bits,
            save_intermediate_results=save_intermediate_results,
            make_output_path=make_output_path,
        )

    # Perform outlier detection using statistical comparisons between
    # each original input image and its blotted version of the median image
    with library:
        for image in library:
            if resample_data:
                _flag_resampled_model_crs(
                    image,
                    median_data,
                    median_wcs,
                    snr1,
                    snr2,
                    scale1,
                    scale2,
                    backg,
                    save_intermediate_results,
                    make_output_path,
                )
            else:
                _flag_model_crs(image, median_data, snr1)

            # mark step as complete
            image.meta.cal_step["outlier_detection"] = "COMPLETE"

            library.shelve(image, modify=True)

    return library



"""
Module to detect sources using image segmentation.
"""

import logging
import math
import warnings

from astropy.convolution import convolve
from astropy.utils.exceptions import AstropyUserWarning
from photutils.segmentation import SourceFinder, make_2dgaussian_kernel
from photutils.utils.exceptions import NoDetectionsWarning

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def convolve_data(data, kernel_fwhm, size=None, mask=None):
    """
    Convolve the background-subtracted model image with a Gaussian
    kernel.

    Parameters
    ----------
    data : 2D `numpy.ndarray`
        The input 2D array. The data is assumed to be
        background subtracted.
    kernel_fwhm : float
        The FWHM of the Gaussian kernel.
    size : int, optional
        The size of the kernel array. Default is 3 times the kernel
        FWHM.
    mask : 2D `numpy.ndarray`, optional
        A boolean mask with the same shape as the input data. True
        values indicate masked pixels.

    Returns
    -------
    convolved_data : `numpy.ndarray`
        The convolved data array.
    """
    size = math.ceil(kernel_fwhm * 3)
    size = size + 1 if size % 2 == 0 else size  # make size be odd
    kernel = make_2dgaussian_kernel(kernel_fwhm, size=size)  # normalized to 1

    with warnings.catch_warnings():
        # suppress warnings caused by large masked areas
        warnings.simplefilter("ignore", AstropyUserWarning)
        return convolve(data, kernel, mask=mask)


def make_segmentation_image(
    convolved_data, snr_threshold, npixels, bkg_rms, deblend=False, mask=None
):
    """
    Make a segmentation image from a model image.

    Parameters
    ----------
    convolved_data : 2D `numpy.ndarray`
        The background-subtracted convolved data array.
    snr_threshold : float
        The per-pixel signal-to-noise ratio threshold for detection.
    npixels : int
        The number of connected pixels required to define a source.
    bkg_rms : 2D `numpy.ndarray`
        The background RMS array.
    deblend : bool, optional
        Whether to deblend overlapping sources. Default is False.
    mask : 2D `numpy.ndarray`, optional
        A boolean mask with the same shape as the input data. True
        values indicate masked pixels.

    Returns
    -------
    segment_img : `SegmentationImage`
        The segmentation image.
    """
    with warnings.catch_warnings():
        # suppress NoDetectionsWarning from photutils
        warnings.filterwarnings("ignore", category=NoDetectionsWarning)

        finder = SourceFinder(npixels, deblend=deblend)
        threshold = snr_threshold * bkg_rms
        segment_img = finder(convolved_data, threshold, mask=mask)

        if segment_img is None:
            nsources = 0
        else:
            nsources = segment_img.nlabels
        log.info(f"Detected {nsources} sources")

    return segment_img



"""Tests for asn_from_list"""

import pytest

from romancal.associations import Association, load_asn
from romancal.associations.asn_from_list import asn_from_list
from romancal.associations.exceptions import AssociationNotValidError


def test_base_association():
    """Create the simplest of associations"""
    items = ["a", "b", "c"]
    asn = asn_from_list(items, rule=Association)
    assert asn["asn_rule"] == "Association"
    assert asn["asn_type"] == "None"
    assert asn["members"] == items


def test_base_roundtrip():
    """Write/read created base association"""
    items = ["a", "b", "c"]
    asn = asn_from_list(items, rule=Association)
    _, serialized = asn.dump()
    reloaded = load_asn(serialized, registry=None)
    assert asn["asn_rule"] == reloaded["asn_rule"]
    assert asn["asn_type"] == reloaded["asn_type"]
    assert asn["members"] == reloaded["members"]


def test_default_simple():
    """Default Level3 association"""
    product_name = "test_product"
    items = ["a", "b", "c"]
    asn = asn_from_list(items, product_name=product_name)
    assert asn["asn_rule"] == "DMS_ELPP_Base"
    assert asn["asn_type"] == "None"
    assert len(asn["products"]) == 1
    product = asn["products"][0]
    assert product["name"] == product_name
    assert len(product["members"]) == len(items)
    for member in product["members"]:
        assert member["expname"] in items
        assert member["exptype"] == "science"


def test_default_with_type():
    """ELPP association with types specified"""
    product_name = "test_product"
    items = {"a": "science", "b": "target_acq", "c": "somethingelse"}
    asn = asn_from_list(
        [(item, type_) for item, type_ in items.items()],
        product_name=product_name,
        with_exptype=True,
    )
    assert asn["asn_rule"] == "DMS_ELPP_Base"
    assert asn["asn_type"] == "None"
    assert len(asn["products"]) == 1
    product = asn["products"][0]
    assert product["name"] == product_name
    assert len(product["members"]) == len(items)
    for member in product["members"]:
        assert member["expname"] in items
        assert member["exptype"] == items[member["expname"]]


def test_default_fail():
    """Test default DMS_ELPP_Base fail

    A product name needs to be included, but is not.
    """
    items = ["a"]
    with pytest.raises(AssociationNotValidError):
        _ = asn_from_list(items)



import pytest
import roman_datamodels as rdm

from romancal.pipeline.mosaic_pipeline import MosaicPipeline

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(scope="module")
def run_mos(rtdata_module):
    rtdata = rtdata_module

    # Test Pipeline
    rtdata.get_asn("WFI/image/L3_mosaic_asn.json")
    output = "r0099101001001001001_r274dp63x31y81_prompt_F158_i2d.asdf"
    rtdata.output = output
    args = [
        "roman_mos",
        rtdata.input,
    ]
    MosaicPipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_mos):
    return run_mos.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_mos):
    return run_mos.truth


def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_resample_ran(output_model):
    # DMS373 test output is resampled to a skycell
    # FIXME this doesn't test if the output is a skycell
    assert output_model.meta.cal_step.resample == "COMPLETE"


def test_location_name(output_model):
    # test that the location_name matches the skycell selected
    assert output_model.meta.basic.location_name == "r274dp63x31y81"



"""Regression test for the Jump detection step."""

import pytest

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
@pytest.mark.skip(reason="Jump detection step is being removed from the pipeline")
def test_jump_detection_step(rtdata, ignore_asdf_paths):
    """Function to run and compare Jump Detection files. Note: This should
    include tests for overrides etc."""

    input_file = "r0000101001001001001_0001_wfi01_darkcurrent.asdf"
    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file

    # Note: the thresholds should be reset to the defaults once we have better
    # input data
    args = [
        "romancal.step.JumpStep",
        rtdata.input,
        "--rejection_threshold=180.",
        "--three_group_rejection_threshold=185.",
        "--four_group_rejection_threshold=190.",
        # disable the jump detection in the rampfit step so this one will run
        "--run_ramp_jump_detection=False",
    ]
    RomanStep.from_cmdline(args)
    output = "r0000101001001001001_0001_wfi01_jump.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()



"""Tests for skycell_asn"""

import pytest

# from romancal.associations import Association, AssociationRegistry, load_asn
import romancal.associations.skycell_asn as skycell_asn
from romancal.associations.skycell_asn import Main


def test_cmdline_fails():
    """Exercise the command line interface"""

    # No arguments
    with pytest.raises(SystemExit):
        Main([])

    # Only the association file argument
    with pytest.raises(SystemExit):
        Main(["-o", "test_asn.json"])


def test_parse_visitID():

    filelist1 = [
        "r0000101002003004005_0001_wfi10_cal.asdf",
    ]
    visitid_parts = skycell_asn.parse_visitID(filelist1[0][1:20])
    assert visitid_parts["Program"] == "00001"
    assert visitid_parts["Execution"] == "01"
    assert visitid_parts["Pass"] == "002"
    assert visitid_parts["Segment"] == "003"
    assert visitid_parts["Observation"] == "004"
    assert visitid_parts["Visit"] == "005"



"""Test basic usage of Level2 associations"""

from romancal.associations import generate, load_asn
from romancal.associations.main import Main
from romancal.associations.tests.helpers import (
    combine_pools,
    registry_level2_only,
    t_path,
)

# REGEX_LEVEL2 = r'(?P<path>.+)(?P<type>_rate(ints)?)(?P<extension>\..+)'
# REGEX_LEVEL2 = r'(?P<path>.+)(.*\_.*\_.*\_.*\_.*)'
REGEX_LEVEL2 = r"(?P<path>.+)(?P<type>_cal?)(?P<extension>\..+)"


def from_level2_schema():
    with open(t_path("data/asn_level2.json")) as asn_file:
        asn = load_asn(asn_file)
    return [asn]


def generate_from_pool(pool_path):
    """Generate associations from pools"""
    rules = registry_level2_only()
    pool = combine_pools(t_path(pool_path))
    asns = generate(pool, rules)
    return asns


def cmd_from_pool(pool_path, args):
    """Run commandline on pool

    Parameters
    ---------
    pool_path: str
        The pool to run on.

    args: [arg(, ...)]
        Additional command line arguments in the form `sys.argv`
    """
    full_args = [
        "--dry-run",
        "-D",
        "-r",
        t_path("../lib/rules_level2.py"),
        "--ignore-default",
    ]
    full_args.extend(args)
    result = Main(full_args, pool=pool_path)
    return result


def test_level2_productname():
    asns = generate_from_pool("data/pool_002_wfi_image.csv")
    for asn in asns:
        for product in asn["products"]:
            science = [
                member
                for member in product["members"]
                if member["exptype"] == "science" or member["exptype"] == "wfi_image"
            ]
            if asn["asn_rule"] == "Asn_Lv2Image":
                assert len(science) == 2
            if asn["asn_rule"] == "Asn_Lv2FOV":
                assert len(science) == 18


#            match = re.match(REGEX_LEVEL2, science[0]['expname'])
#            pdb.set_trace()
#            assert match.groupdict()['path'] == product['name']



import os

import numpy as np
import pytest

from romancal.datamodels import ModelLibrary
from romancal.outlier_detection import OutlierDetectionStep


@pytest.mark.parametrize(
    "input_models",
    [
        "",
    ],
)
def test_outlier_raises_error_on_invalid_input_models(input_models):
    """Test that OutlierDetection raises an Exception if input is invalid."""

    with pytest.raises(IsADirectoryError):
        OutlierDetectionStep.call(input_models)


def test_outlier_skips_step_on_invalid_number_of_elements_in_input(base_image):
    """Test that OutlierDetection skips processing when provided with an invalid number of elements in the input,
    and sets the appropriate metadata for the skipped step."""
    img = base_image()

    res = OutlierDetectionStep.call(ModelLibrary([img]))

    with res:
        for i, m in enumerate(res):
            assert m.meta.cal_step.outlier_detection == "SKIPPED"
            res.shelve(m, i, modify=False)


def test_outlier_raises_exception_on_exposure_type_different_from_wfi_image(base_image):
    """
    Test if the outlier detection step raises an Exception for non-image inputs.
    """
    img_1 = base_image()
    img_1.meta.exposure.type = "WFI_PRISM"
    img_2 = base_image()
    img_2.meta.exposure.type = "WFI_PRISM"

    with pytest.raises(ValueError, match="only supports WFI_IMAGE"):
        OutlierDetectionStep.call(ModelLibrary([img_1, img_2]))


def test_outlier_valid_input_asn(tmp_path, base_image, create_mock_asn_file):
    """
    Test that OutlierDetection runs with valid ASN file as input.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_1.save(tmp_path / "img_1.asdf")
    img_2 = base_image()
    img_1.meta.filename = "img_2.asdf"
    img_2.save(tmp_path / "img_2.asdf")

    asn_filepath = create_mock_asn_file(tmp_path)

    res = OutlierDetectionStep.call(
        asn_filepath,
        output_dir=tmp_path.as_posix(),
        in_memory=True,
        resample_data=False,
    )

    # assert step.skip is False
    with res:
        for i, m in enumerate(res):
            assert m.meta.cal_step.outlier_detection == "COMPLETE"
            res.shelve(m, i, modify=False)


def test_outlier_valid_input_modelcontainer(tmp_path, base_image):
    """
    Test that OutlierDetection runs with valid ModelLibrary as input.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"

    library = ModelLibrary([img_1, img_2])

    res = OutlierDetectionStep.call(
        library,
        in_memory=True,
        resample_data=False,
    )

    with res:
        for i, m in enumerate(res):
            assert m.meta.cal_step.outlier_detection == "COMPLETE"
            res.shelve(m, i, modify=False)


def test_outlier_do_detection_write_files_to_custom_location(tmp_path, base_image):
    """
    Test that OutlierDetection can create files on disk in a custom location.
    """
    img_1 = base_image()
    img_1.meta.filename = "img1_cal.asdf"
    img_1.meta.background.level = 0
    img_2 = base_image()
    img_2.meta.filename = "img2_cal.asdf"
    img_2.meta.background.level = 0
    input_models = ModelLibrary([img_1, img_2])

    outlier_step = OutlierDetectionStep(
        in_memory=False,
        save_intermediate_results=True,
    )
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path.as_posix()

    outlier_step(input_models)

    # meta.filename for the median image created by OutlierDetection.do_detection()
    median_path = tmp_path / "drizzled_median.asdf"

    outlier_files_path = [
        median_path,
    ]

    assert all(x.exists() for x in outlier_files_path)


@pytest.mark.parametrize("on_disk", (True, False))
def test_find_outliers(tmp_path, base_image, on_disk):
    """
    Test that OutlierDetection can find outliers.
    """
    cr_value = 100
    source_value = 10
    err_value = 10  # snr=1

    imgs = []
    for i in range(3):
        img = base_image()
        img.data[42, 72] = source_value
        img.err[:] = err_value
        img.meta.filename = str(tmp_path / f"img{i}_suffix.asdf")
        img.meta.observation.observation_id = str(i)
        img.meta.background.level = 0
        imgs.append(img)

    # add outliers
    img_0_input_coords = np.array(
        [[5, 25, 45, 65, 85], [45, 25, 85, 65, 5]], dtype=np.int64
    )
    img_1_input_coords = np.array(
        [[15, 35, 75, 95, 99], [25, 5, 65, 45, 5]], dtype=np.int64
    )

    imgs[0].data[img_0_input_coords[0], img_0_input_coords[1]] = cr_value
    imgs[1].data[img_1_input_coords[0], img_1_input_coords[1]] = cr_value

    if on_disk:
        # write out models and asn to disk
        for img in imgs:
            img.save(img.meta.filename)
        asn_dict = {
            "asn_id": "a3001",
            "asn_pool": "none",
            "products": [
                {
                    "name": "test_asn",
                    "members": [
                        {"expname": m.meta.filename, "exptype": "science"} for m in imgs
                    ],
                }
            ],
        }
        input_models = ModelLibrary(asn_dict, on_disk=True)
    else:
        input_models = ModelLibrary(imgs)

    outlier_step = OutlierDetectionStep()
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path.as_posix()
    # make sure files are written out to disk
    outlier_step.in_memory = not on_disk

    result = outlier_step(input_models)

    expected_crs = [img_0_input_coords, img_1_input_coords, None]
    with result:
        for cr_coords, flagged_img in zip(expected_crs, result):
            if cr_coords is None:
                assert not np.any(flagged_img.dq > 0)
            else:
                flagged_coords = np.where(flagged_img.dq > 0)
                np.testing.assert_equal(cr_coords, flagged_coords)
            result.shelve(flagged_img, modify=False)


def test_identical_images(tmp_path, base_image, caplog):
    """
    Test that OutlierDetection does not flag any outliers in the DQ array if images are identical.
    """
    img_1 = base_image()
    img_1.meta.filename = "img1_suffix.asdf"
    img_1.meta.background.level = 0
    # add outliers
    img_1_input_coords = np.array(
        [(5, 45), (25, 25), (45, 85), (65, 65), (85, 5)], dtype=[("x", int), ("y", int)]
    )
    img_1.data[img_1_input_coords["x"], img_1_input_coords["y"]] = 100000

    img_2 = img_1.copy()
    img_2.meta.filename = "img2_suffix.asdf"
    img_3 = img_1.copy()
    img_3.meta.filename = "img3_suffix.asdf"

    input_models = ModelLibrary([img_1, img_2, img_3])

    outlier_step = OutlierDetectionStep()
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path.as_posix()
    # make sure files are written out to disk
    outlier_step.in_memory = False

    result = outlier_step(input_models)

    # assert that log shows no new outliers detected
    assert "0 pixels marked as outliers" in {x.message for x in caplog.records}
    # assert that DQ array has nothing flagged as outliers
    with result:
        for i, model in enumerate(result):
            assert np.count_nonzero(model.dq) == 0
            result.shelve(model, i)


@pytest.mark.parametrize(
    "input_type",
    [
        "ModelLibrary",
        "ASNFile",
        "DataModelList",
        "ASDFFilenameList",
    ],
)
def test_outlier_detection_always_returns_modelcontainer_with_updated_datamodels(
    input_type,
    base_image,
    tmp_path,
    create_mock_asn_file,
):
    """Test that the OutlierDetectionStep always returns a ModelLibrary
    with updated data models after processing different input types."""

    os.chdir(tmp_path)
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_1.data *= img_1.meta.photometry.conversion_megajanskys / img_1.data
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"
    img_2.data *= img_2.meta.photometry.conversion_megajanskys / img_2.data

    library = ModelLibrary([img_1, img_2])
    library._save(tmp_path)

    step_input_map = {
        "ModelLibrary": library,
        "ASNFile": create_mock_asn_file(
            tmp_path,
            members_mapping=[
                {"expname": img_1.meta.filename, "exptype": "science"},
                {"expname": img_2.meta.filename, "exptype": "science"},
            ],
        ),
        "DataModelList": [img_1, img_2],
        "ASDFFilenameList": [img_1.meta.filename, img_2.meta.filename],
    }

    step_input = step_input_map.get(input_type)

    res = OutlierDetectionStep.call(step_input)

    assert isinstance(res, ModelLibrary)
    with res:
        for i, model in enumerate(res):
            assert model.meta.cal_step.outlier_detection == "COMPLETE"
            res.shelve(model, i, modify=False)



"""
Entry point implementations.
"""


def get_steps():
    """
    Return tuples describing the stpipe.Step subclasses provided
    by this package.  This method is registered with the stpipe.steps
    entry point.

    Returns
    -------
    list of tuple (str, str, bool)
        The first element each tuple is a fully-qualified Step
        subclass name.  The second element is an optional class
        alias.  The third element indicates that the class
        is a subclass of Pipeline.
    """
    # Unit tests ensure that this list is kept in sync with the actual
    # class definitions.  We need to avoid importing romancal.pipeline and
    # romancal.step to keep the CLI snappy.
    return [
        ("romancal.pipeline.ExposurePipeline", "roman_elp", True),
        ("romancal.pipeline.MosaicPipeline", "roman_mos", True),
        ("romancal.step.DarkCurrentStep", "dark", False),
        ("romancal.step.DQInitStep", "dq_init", False),
        ("romancal.step.FlatFieldStep", "flat_field", False),
        ("romancal.step.FluxStep", "flux", False),
        ("romancal.step.JumpStep", "jump", False),
        ("romancal.step.LinearityStep", "linearity", False),
        ("romancal.step.PhotomStep", "photom", False),
        ("romancal.step.RampFitStep", "ramp_fit", False),
        ("romancal.step.RefPixStep", "refpix", False),
        ("romancal.step.SaturationStep", "saturation", False),
        ("romancal.step.AssignWcsStep", "assign_wcs", False),
        ("romancal.step.OutlierDetectionStep", "outlier_detection", False),
        ("romancal.step.SourceDetectionStep", "source_detection", False),
        ("romancal.step.SkyMatchStep", "skymatch", False),
        ("romancal.step.TweakRegStep", "tweakreg", False),
        ("romancal.step.ResampleStep", "resample", False),
        ("romancal.step.SourceCatalogStep", "source_catalog", False),
        ("romancal.step.MultibandCatalogStep", "multiband_catalog", False),
    ]



"""
This module collects all of the stpipe.Step subclasses
made available by this package.
"""

from .assign_wcs.assign_wcs_step import AssignWcsStep
from .dark_current.dark_current_step import DarkCurrentStep
from .dq_init.dq_init_step import DQInitStep
from .flatfield.flat_field_step import FlatFieldStep
from .flux import FluxStep
from .jump.jump_step import JumpStep
from .linearity.linearity_step import LinearityStep
from .multiband_catalog.multiband_catalog_step import MultibandCatalogStep
from .outlier_detection.outlier_detection_step import OutlierDetectionStep
from .photom.photom_step import PhotomStep
from .ramp_fitting.ramp_fit_step import RampFitStep
from .refpix.refpix_step import RefPixStep
from .resample.resample_step import ResampleStep
from .saturation.saturation_step import SaturationStep
from .skymatch.skymatch_step import SkyMatchStep
from .source_catalog.source_catalog_step import SourceCatalogStep
from .source_detection.source_detection_step import SourceDetectionStep
from .tweakreg.tweakreg_step import TweakRegStep

__all__ = [
    "AssignWcsStep",
    "DarkCurrentStep",
    "DQInitStep",
    "FlatFieldStep",
    "FluxStep",
    "JumpStep",
    "LinearityStep",
    "OutlierDetectionStep",
    "PhotomStep",
    "RampFitStep",
    "RefPixStep",
    "ResampleStep",
    "SaturationStep",
    "SkyMatchStep",
    "SourceDetectionStep",
    "SourceCatalogStep",
    "MultibandCatalogStep",
    "TweakRegStep",
]



"""Association Definitions: DMS-specific

Notes
-----
These associations are specifically defined for use in DMS.
"""

from romancal.associations.lib import rules_level2
from romancal.associations.registry import RegistryMarker

RegistryMarker.mark(rules_level2)



import json
import re
from pathlib import Path

TEST_DIRECTORY = Path(__file__).parent.parent
TEST_REQUIREMENTS_FILENAME = TEST_DIRECTORY / "tests" / "dms_requirement_tests.json"


def test_requirements():
    test_requirements_filename = TEST_REQUIREMENTS_FILENAME.expanduser().absolute()
    test_directory = TEST_DIRECTORY.expanduser().absolute()

    with open(test_requirements_filename) as test_requirements_file:
        requirements = json.load(test_requirements_file)

    required_tests = {
        test
        for requirement_tests in requirements.values()
        for test in requirement_tests
    }

    existing_tests = set()
    test_regex = re.compile(r"def (test_[^\(]+)\(")
    for test_filename in test_directory.glob("**/test_*.py"):
        with open(test_filename) as test_file:
            test_file_contents = test_file.read()

        for match in re.finditer(test_regex, test_file_contents):
            test = f"{test_directory.stem}.{str(test_filename.relative_to(test_directory).parent).replace('/', '.')}.{test_filename.stem}.{match.group(1)}"
            if test in required_tests:
                existing_tests.add(test)

    missing_tests = required_tests - existing_tests
    assert (
        not missing_tests
    ), f"could not find the following tests correlated with DMS requirements: {missing_tests}"



import asdf
from roman_datamodels import open as datamodels_open
from stpipe.library import AbstractModelLibrary, NoGroupID

from romancal.associations import AssociationNotValidError, load_asn

__all__ = ["ModelLibrary"]


class ModelLibrary(AbstractModelLibrary):
    @property
    def crds_observatory(self):
        return "roman"

    def _model_to_filename(self, model):
        model_filename = model.meta.filename
        if model_filename is None:
            model_filename = "model.asdf"
        return model_filename

    def _datamodels_open(self, filename, **kwargs):
        return datamodels_open(filename, **kwargs)

    @classmethod
    def _load_asn(cls, asn_path):
        try:
            with open(asn_path) as asn_file:
                asn_data = load_asn(asn_file)
        except AssociationNotValidError as e:
            raise OSError("Cannot read ASN file.") from e
        return asn_data

    def _filename_to_group_id(self, filename):
        """
        Compute a "group_id" without loading the file as a DataModel

        This function will return the meta.group_id stored in the ASDF
        extension (if it exists) or a group_id calculated from the
        ASDF headers.
        """
        meta = asdf.util.load_yaml(filename)["roman"]["meta"]
        if group_id := meta.get("group_id"):
            return group_id
        if "observation" in meta:
            return _mapping_to_group_id(meta["observation"])
        raise NoGroupID(f"{filename} missing group_id")

    def _model_to_group_id(self, model):
        """
        Compute a "group_id" from a model using the DataModel interface
        """
        if (group_id := getattr(model.meta, "group_id", None)) is not None:
            return group_id
        if hasattr(model.meta, "observation"):
            return _mapping_to_group_id(model.meta.observation)
        raise NoGroupID(f"{model} missing group_id")

    def _assign_member_to_model(self, model, member):
        # roman_datamodels doesn't allow assignment of meta.group_id
        # (since it's not in the schema). To work around this use
        # __setitem__ calls here instead of setattr
        for attr in ("group_id", "tweakreg_catalog", "exptype"):
            if attr in member:
                model.meta[attr] = member[attr]
        if not hasattr(model.meta, "asn"):
            model.meta["asn"] = {}

        if "table_name" in self.asn:
            model.meta.asn["table_name"] = self.asn["table_name"]
        if "asn_pool" in self.asn:
            model.meta.asn["pool_name"] = self.asn["asn_pool"]


def _mapping_to_group_id(mapping):
    """
    Combine a number of file metadata values into a ``group_id`` string
    """
    return "{observation_id}".format_map(mapping)



import numpy as np
import pytest
from gwcs.wcstools import grid_from_bounding_box
from numpy.testing import assert_allclose
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils

from romancal.assign_wcs.assign_wcs_step import AssignWcsStep, load_wcs
from romancal.assign_wcs.utils import wcs_bbox_from_shape


def create_image():
    l2 = maker_utils.mk_level2_image()

    l2.meta.wcsinfo.v2_ref = -503
    l2.meta.wcsinfo.v3_ref = -318
    l2.meta.wcsinfo.ra_ref = 156
    l2.meta.wcsinfo.dec_ref = 54.2
    l2.meta.wcsinfo.vparity = -1
    l2.meta.wcsinfo.roll_ref = 0.15

    l2im = rdm.ImageModel(l2)
    return l2im


def create_distortion():
    distortions = [maker_utils.mk_distortion()]

    model = create_image()
    dist = maker_utils.mk_distortion()
    dist.coordinate_distortion_transform.bounding_box = wcs_bbox_from_shape(
        model.data.shape
    )
    distortions.append(dist)

    return distortions


def create_step():
    def load_wcs_step(image, file_name):
        return load_wcs(image, {"distortion": file_name})

    def assign_wcs_step(image, file_name):
        return AssignWcsStep.call(image, override_distortion=file_name)

    return [load_wcs_step, assign_wcs_step]


@pytest.mark.parametrize("distortion", create_distortion())
@pytest.mark.parametrize("step", create_step())
def test_wcs(tmp_path, distortion, step):
    file_name = str(tmp_path / "distortion.asdf")
    dist = rdm.DistortionRefModel(distortion)
    dist.save(file_name)

    l2im = create_image()
    l2_wcs = step(l2im, file_name)

    assert l2_wcs.meta.wcs is not None
    assert l2_wcs.meta.cal_step.assign_wcs == "COMPLETE"

    x, y = grid_from_bounding_box(l2_wcs.meta.wcs.bounding_box)

    ra, dec = l2_wcs.meta.wcs(x, y)
    assert_allclose(ra, l2im.meta.wcsinfo.ra_ref, atol=2.3)
    assert_allclose(dec, l2im.meta.wcsinfo.dec_ref, atol=1.3)

    # verify inputs outside the bounding box return NaNs
    ra, dec = l2_wcs.meta.wcs(-5, 3)
    assert np.isnan(ra)
    assert np.isnan(dec)

    # check S_REGION length and format
    s_region_list = l2_wcs.meta.wcsinfo.s_region.split()
    assert len(s_region_list) == 10
    assert s_region_list[0].lower() == "polygon"
    assert s_region_list[1].lower() == "icrs"

    # check if BBOX is not None
    assert l2_wcs.meta.wcs.bounding_box is not None

    # check if footprint solution for each detector is within 10% of (RA_REF, DEC_REF)
    s_region_alpha_list = [
        float(x) for i, x in enumerate(s_region_list[2:]) if i % 2 == 0
    ]
    s_region_delta_list = [
        float(x) for i, x in enumerate(s_region_list[2:]) if i % 2 != 0
    ]
    assert_allclose(s_region_alpha_list, l2_wcs.meta.wcsinfo.ra_ref, rtol=1e-1, atol=0)
    assert_allclose(s_region_delta_list, l2_wcs.meta.wcsinfo.dec_ref, rtol=1e-1, atol=0)


@pytest.mark.parametrize("step", create_step())
def test_crds_getbestref(step):
    l2im = create_image()
    l2_wcs = AssignWcsStep.call(l2im, override_distortion="N/A")

    assert l2_wcs.meta.cal_step.assign_wcs == "SKIPPED"
